import math

import torch
import torch.nn as nn
from tabpfn.layer import TransformerEncoderLayer
from tabpfn.utils import SeqBN, bool_mask_to_att_mask
from torch import Tensor
from torch.nn import Module, TransformerEncoder


class TransformerModel(nn.Module):
    def __init__(
        self,
        encoder,
        n_out,
        ninp,
        nhead,
        nhid,
        nlayers,
        dropout=0.0,
        style_encoder=None,
        y_encoder=None,
        pos_encoder=None,
        decoder=None,
        input_normalization=False,
        init_method=None,
        pre_norm=False,
        activation="gelu",
        recompute_attn=False,
        num_global_att_tokens=0,
        full_attention=False,
        all_layers_same_init=False,
        efficient_eval_masking=True,
    ):
        super().__init__()
        self.model_type = "Transformer"
        encoder_layer_creator = lambda: TransformerEncoderLayer(
            ninp, nhead, nhid, dropout, activation=activation, pre_norm=pre_norm, recompute_attn=recompute_attn
        )
        self.transformer_encoder = (
            TransformerEncoder(encoder_layer_creator(), nlayers)
            if all_layers_same_init
            else TransformerEncoderDiffInit(encoder_layer_creator, nlayers)
        )
        self.ninp = ninp
        self.encoder = encoder
        self.y_encoder = y_encoder
        self.pos_encoder = pos_encoder
        self.decoder = (
            decoder(ninp, nhid, n_out)
            if decoder is not None
            else nn.Sequential(nn.Linear(ninp, nhid), nn.GELU(), nn.Linear(nhid, n_out))
        )
        self.input_ln = SeqBN(ninp) if input_normalization else None
        self.style_encoder = style_encoder
        self.init_method = init_method
        if num_global_att_tokens is not None:
            assert not full_attention
        self.global_att_embeddings = nn.Embedding(num_global_att_tokens, ninp) if num_global_att_tokens else None
        self.full_attention = full_attention
        self.efficient_eval_masking = efficient_eval_masking
        self.embedding_shape = None
        self.embeds = True

        self.n_out = n_out
        self.nhid = nhid

        self.init_weights()

    def __setstate__(self, state):
        super().__setstate__(state)
        self.__dict__.setdefault("efficient_eval_masking", False)

    @staticmethod
    def generate_square_subsequent_mask(sz):
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        return bool_mask_to_att_mask(mask)

    @staticmethod
    def generate_D_q_matrix(sz, query_size):
        train_size = sz - query_size
        mask = torch.zeros(sz, sz) == 0
        mask[:, train_size:].zero_()
        mask |= torch.eye(sz) == 1
        return bool_mask_to_att_mask(mask)

    @staticmethod
    def generate_global_att_query_matrix(num_global_att_tokens, seq_len, num_query_tokens):
        train_size = seq_len + num_global_att_tokens - num_query_tokens
        sz = seq_len + num_global_att_tokens
        mask = torch.zeros(num_query_tokens, sz) == 0
        mask[:, train_size:].zero_()
        mask[:, train_size:] |= torch.eye(num_query_tokens) == 1
        return bool_mask_to_att_mask(mask)

    @staticmethod
    def generate_global_att_trainset_matrix(num_global_att_tokens, seq_len, num_query_tokens):
        train_size = seq_len + num_global_att_tokens - num_query_tokens
        trainset_size = seq_len - num_query_tokens
        mask = torch.zeros(trainset_size, num_global_att_tokens) == 0
        # mask[:,num_global_att_tokens:].zero_()
        # mask[:,num_global_att_tokens:] |= torch.eye(trainset_size) == 1
        return bool_mask_to_att_mask(mask)

    @staticmethod
    def generate_global_att_globaltokens_matrix(num_global_att_tokens, seq_len, num_query_tokens):
        mask = torch.zeros(num_global_att_tokens, num_global_att_tokens + seq_len - num_query_tokens) == 0
        return bool_mask_to_att_mask(mask)

    def init_weights(self):
        # print("INIT")
        initrange = 1.0
        # if isinstance(self.encoder,EmbeddingEncoder):
        #    self.encoder.weight.data.uniform_(-initrange, initrange)
        # self.decoder.bias.data.zero_()
        # self.decoder.weight.data.uniform_(-initrange, initrange)
        if self.init_method is not None:
            self.apply(self.init_method)
        for layer in self.transformer_encoder.layers:
            nn.init.zeros_(layer.linear2.weight)
            nn.init.zeros_(layer.linear2.bias)
            attns = layer.self_attn if isinstance(layer.self_attn, nn.ModuleList) else [layer.self_attn]
            for attn in attns:
                nn.init.zeros_(attn.out_proj.weight)
                nn.init.zeros_(attn.out_proj.bias)

    def forward(self, src, src_mask=None, single_eval_pos=None):
        assert isinstance(src, tuple), "inputs (src) have to be given as (x,y) or (style,x,y) tuple"
        # print("SOURCE MASK",src_mask)
        if len(src) == 2:  # (x,y) and no style
            src = (None,) + src

        style_src, x_src, y_src = src
        x_src = self.encoder(x_src)
        y_src = self.y_encoder(y_src.unsqueeze(-1) if len(y_src.shape) < len(x_src.shape) else y_src)
        style_src = (
            self.style_encoder(style_src).unsqueeze(0) if self.style_encoder else torch.tensor([], device=x_src.device)
        )
        global_src = (
            torch.tensor([], device=x_src.device)
            if self.global_att_embeddings is None
            else self.global_att_embeddings.weight.unsqueeze(1).repeat(1, x_src.shape[1], 1)
        )

        if src_mask is not None:
            assert self.global_att_embeddings is None or isinstance(src_mask, tuple)
        if src_mask is None:
            if self.global_att_embeddings is None:
                full_len = len(x_src) + len(style_src)
                if self.full_attention:
                    src_mask = bool_mask_to_att_mask(torch.ones((full_len, full_len), dtype=torch.bool)).to(
                        x_src.device
                    )
                elif self.efficient_eval_masking:
                    src_mask = single_eval_pos + len(style_src)
                else:
                    src_mask = self.generate_D_q_matrix(full_len, len(x_src) - single_eval_pos).to(x_src.device)
            else:
                src_mask_args = (
                    self.global_att_embeddings.num_embeddings,
                    len(x_src) + len(style_src),
                    len(x_src) + len(style_src) - single_eval_pos,
                )
                src_mask = (
                    self.generate_global_att_globaltokens_matrix(*src_mask_args).to(x_src.device),
                    self.generate_global_att_trainset_matrix(*src_mask_args).to(x_src.device),
                    self.generate_global_att_query_matrix(*src_mask_args).to(x_src.device),
                )
        # print(x_src[:single_eval_pos].size())
        train_x = x_src[:single_eval_pos] + y_src[:single_eval_pos]
        # print(train_x.size())
        # print(y_src[:single_eval_pos].size())
        src = torch.cat([global_src, style_src, train_x, x_src[single_eval_pos:]], 0)

        if self.input_ln is not None:
            src = self.input_ln(src)

        if self.pos_encoder is not None:
            src = self.pos_encoder(src)

        output = self.transformer_encoder(src, src_mask)
        # take only the "test" because they only depend on the features, not on the labels
        embeds = output[single_eval_pos:].detach().cpu().numpy()
        # print("EMEBD", embeds)

        # embeds=np.mean(embeds,1)
        embeds = embeds.reshape(embeds.shape[0], embeds.shape[1] * embeds.shape[2])


        self.embedding_shape = embeds.shape
        # pca = PCA(n_components=10)
        # embeds = pca.fit_transform(embeds)

        """
        if self.embeds:
            print("train embeddings saved")
            np.save("embeds_train.npy",embeds)
        if not self.embeds:
            print("test embeddings saved")
            np.save("embeds_test.npy",embeds)
        """

        return torch.from_numpy(embeds)

    @torch.no_grad()
    def init_from_small_model(self, small_model):
        assert (
            isinstance(self.decoder, nn.Linear)
            and isinstance(self.encoder, (nn.Linear, nn.Sequential))
            and isinstance(self.y_encoder, (nn.Linear, nn.Sequential))
        )

        def set_encoder_weights(my_encoder, small_model_encoder):
            my_encoder_linear, small_encoder_linear = (
                (my_encoder, small_model_encoder)
                if isinstance(my_encoder, nn.Linear)
                else (my_encoder[-1], small_model_encoder[-1])
            )
            small_in_dim = small_encoder_linear.out_features
            my_encoder_linear.weight.zero_()
            my_encoder_linear.bias.zero_()
            my_encoder_linear.weight[:small_in_dim] = small_encoder_linear.weight
            my_encoder_linear.bias[:small_in_dim] = small_encoder_linear.bias

        set_encoder_weights(self.encoder, small_model.encoder)
        set_encoder_weights(self.y_encoder, small_model.y_encoder)

        small_in_dim = small_model.decoder.in_features

        self.decoder.weight[:, :small_in_dim] = small_model.decoder.weight
        self.decoder.bias = small_model.decoder.bias

        for my_layer, small_layer in zip(self.transformer_encoder.layers, small_model.transformer_encoder.layers):
            small_hid_dim = small_layer.linear1.out_features
            my_in_dim = my_layer.linear1.in_features

            # packed along q,k,v order in first dim
            my_in_proj_w = my_layer.self_attn.in_proj_weight
            small_in_proj_w = small_layer.self_attn.in_proj_weight

            my_in_proj_w.view(3, my_in_dim, my_in_dim)[:, :small_in_dim, :small_in_dim] = small_in_proj_w.view(
                3, small_in_dim, small_in_dim
            )
            my_layer.self_attn.in_proj_bias.view(3, my_in_dim)[:, :small_in_dim] = (
                small_layer.self_attn.in_proj_bias.view(3, small_in_dim)
            )

            my_layer.self_attn.out_proj.weight[:small_in_dim, :small_in_dim] = small_layer.self_attn.out_proj.weight
            my_layer.self_attn.out_proj.bias[:small_in_dim] = small_layer.self_attn.out_proj.bias

            my_layer.linear1.weight[:small_hid_dim, :small_in_dim] = small_layer.linear1.weight
            my_layer.linear1.bias[:small_hid_dim] = small_layer.linear1.bias

            my_layer.linear2.weight[:small_in_dim, :small_hid_dim] = small_layer.linear2.weight
            my_layer.linear2.bias[:small_in_dim] = small_layer.linear2.bias

            my_layer.norm1.weight[:small_in_dim] = math.sqrt(small_in_dim / my_in_dim) * small_layer.norm1.weight
            my_layer.norm2.weight[:small_in_dim] = math.sqrt(small_in_dim / my_in_dim) * small_layer.norm2.weight

            my_layer.norm1.bias[:small_in_dim] = small_layer.norm1.bias
            my_layer.norm2.bias[:small_in_dim] = small_layer.norm2.bias


class TransformerEncoderDiffInit(Module):
    r"""TransformerEncoder is a stack of N encoder layers

    Args:
        encoder_layer_creator: a function generating objects of TransformerEncoderLayer class without args (required).
        num_layers: the number of sub-encoder-layers in the encoder (required).
        norm: the layer normalization component (optional).
    """

    __constants__ = ["norm"]

    def __init__(self, encoder_layer_creator, num_layers, norm=None):
        super().__init__()
        self.layers = nn.ModuleList([encoder_layer_creator() for _ in range(num_layers)])
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self, src: Tensor, mask: Tensor | None = None, src_key_padding_mask: Tensor | None = None
    ) -> Tensor:
        r"""Pass the input through the encoder layers in turn.

        Args:
            src: the sequence to the encoder (required).
            mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        output = src

        for mod in self.layers:
            output = mod(output, src_mask=mask, src_key_padding_mask=src_key_padding_mask)

        if self.norm is not None:
            output = self.norm(output)

        return output



import warnings
from abc import ABC, abstractmethod

import numpy as np
from skactiveml.classifier import SklearnClassifier
from skactiveml.pool import (
    BatchBALD,
    DiscriminativeAL,
    EpistemicUncertaintySampling,
    MonteCarloEER,
    QueryByCommittee,
    RandomSampling,
    UncertaintySampling,
)
from sklearn.cluster import AgglomerativeClustering, KMeans

from alpbench.util.common import fullname
from alpbench.util.ensemble_constructor import Ensemble as Ens


class QueryStrategy(ABC):
    """QueryStrategy

    This class is an abstract class for query strategies. The query strategies are used to sample instances from the
    pool of unlabeled instances.
    """

    @abstractmethod
    def sample(self, learner, X_l, y_l, X_u, num_queries):
        """
        This method samples instances from the pool of unlabeled instances. It is given a learner, that
        is already fitted on the labeled data and potentially used to predict probabilities for the unlabeled
        data.

        Parameters:
            learner: object
            X_l: np.ndarray
            y_l: np.ndarray
            X_u: np.ndarray
            num_queries: int
        """
        pass

    @abstractmethod
    def get_params(self):
        """
        This method returns the parameters of the query strategy.
        """
        pass


class PseudoRandomizedQueryStrategy(QueryStrategy):
    """PseudoRandomizedQueryStrategy

    This class is an abstract class for query strategies that are pseudo-randomized, meaning that they can be
    reproduced with the same random seed.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        self.seed = seed

    @abstractmethod
    def sample(self, learner, X_l, y_l, X_u, num_queries):
        pass

    def get_params(self):
        return {"seed": self.seed}


class WrappedQueryStrategy(QueryStrategy):
    """WrappedQueryStrategy

    This class is used to wrap a query strategy with a learner. The wrapped query strategy is used to sample instances
    from the pool of unlabeled instances.

    Args:
        wrapped_query_strategy: object
        learner: object

    Attributes:
        wrapped_query_strategy: object
        learner: object
    """

    def __init__(self, wrapped_query_strategy: QueryStrategy, learner):
        self.wrapped_query_strategy = wrapped_query_strategy
        self.learner = learner

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        self.learner.fit(X_l, y_l)
        self.wrapped_query_strategy.sample(self.learner, X_l, y_l, X_u, num_queries)

    def get_params(self):
        return {
            "wrapped_query_strategy": {
                "fqn": fullname(self.wrapped_query_strategy),
                "params": self.wrapped_query_strategy.get_params(),
            },
            "learner": {"fqn": fullname(self.learner), "params": self.learner.get_params()},
        }


class ActiveMLQueryStrategy(PseudoRandomizedQueryStrategy):
    """ActiveMLQueryStrategy

    This class is an abstract class for active learning query strategies. The query strategies are used to sample
    instances from the pool of unlabeled instances.

    Args:
        seed (int): The seed for the random number generator.
        qs: object

    Attributes:
        seed (int): The seed for the random number generator.
        qs: object
    """

    def __init__(self, seed, qs):
        super().__init__(seed=seed)
        self.qs = qs

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        nan_labels = np.full(len(X_u), np.nan, dtype=float)
        queried_ids = self.qs.query(
            X=np.concatenate([X_l, X_u]), y=np.concatenate([y_l, nan_labels]), batch_size=num_queries
        )
        queried_original_ids = queried_ids - len(y_l)
        return queried_original_ids


class ActiveMLModelBasedQueryStrategy(ActiveMLQueryStrategy):
    """ActiveMLModelBasedQueryStrategy

    This class is an abstract class for active learning query strategies that are model-based. The query strategies are
    used to sample instances from the pool of unlabeled instances.

    Args:
        seed (int): The seed for the random number generator.
        qs: object

    Attributes:
        seed (int): The seed for the random number generator.
        qs: object
    """

    def __init__(self, seed, qs):
        super().__init__(seed=seed, qs=qs)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        nan_labels = np.full(len(X_u), np.nan, dtype=float)
        queried_ids = self.qs.query(
            clf=SklearnClassifier(learner),
            X=np.concatenate([X_l, X_u]),
            y=np.concatenate([y_l, nan_labels]),
            batch_size=num_queries,
        )
        queried_original_ids = queried_ids - len(y_l)

        return queried_original_ids


class RandomQueryStrategy(ActiveMLQueryStrategy):
    """RandomQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances randomly.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed, RandomSampling(random_state=seed))


class UncertaintyQueryStrategy(ActiveMLQueryStrategy):
    """UncertaintyQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on uncertainty.

    Args:
        seed (int): The seed for the random number generator.
        method (str): The method for uncertainty sampling.

    Attributes:
        seed (int): The seed for the random number generator.
        qs: object
    """

    def __init__(self, seed, method):
        super().__init__(seed=seed, qs=UncertaintySampling(method=method, random_state=seed))


class ExpectedAveragePrecision(UncertaintyQueryStrategy):
    """ExpectedAveragePrecision

    This class is used to sample instances from the pool of unlabeled instances based on the expected average precision.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed, method="expected_average_precision")


class EpistemicUncertaintyQueryStrategy(ActiveMLModelBasedQueryStrategy):
    """EpistemicUncertaintyQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on epistemic uncertainty.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed, EpistemicUncertaintySampling(random_state=seed))


class MonteCarloEERStrategy(ActiveMLModelBasedQueryStrategy):
    """MonteCarloEERStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the Monte Carlo EER method.

    Args:
        seed (int): The seed for the random number generator.
        method (str): The method for Monte Carlo EER.

    Attributes:
        seed (int): The seed for the random number generator.
        qs: object
    """

    def __init__(self, seed, method):
        super().__init__(seed, MonteCarloEER(method=method, random_state=seed))


class MonteCarloEERLogLoss(MonteCarloEERStrategy):
    """MonteCarloEERLogLoss

    This class is used to sample instances from the pool of unlabeled instances based on the Monte Carlo EER method with
    the log loss method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed, method="log_loss")


class MonteCarloEERMisclassification(MonteCarloEERStrategy):
    """MonteCarloEERMisclassification

    This class is used to sample instances from the pool of unlabeled instances based on the Monte Carlo EER method with
    the misclassification loss method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed, method="misclassification_loss")


class DiscriminativeQueryStrategy(ActiveMLQueryStrategy):
    def __init__(self, seed):
        super().__init__(seed=seed, qs=DiscriminativeAL(random_state=seed))

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        nan_labels = np.full(len(X_u), np.nan, dtype=float)
        queried_ids = self.qs.query(
            X=np.concatenate([X_l, X_u]),
            y=np.concatenate([y_l, nan_labels]),
            discriminator=SklearnClassifier(learner),
            batch_size=num_queries,
        )
        queried_original_ids = queried_ids - len(y_l)
        return queried_original_ids


class ActiveMLEnsembleQueryStrategy(ActiveMLQueryStrategy):
    """ActiveMLEnsembleQueryStrategy

    This class is an abstract class for active learning query strategies that are ensemble-based. The query strategies
    are used to sample instances from the pool of unlabeled instances.

    Args:
        seed (int): The seed for the random number generator.
        qs: object
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        qs: object
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, qs, ensemble_size):
        super().__init__(seed=seed, qs=qs)
        self.ensemble_size = ensemble_size

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        learners = [SklearnClassifier(learner)] * self.ensemble_size
        nan_labels = np.full(len(X_u), np.nan, dtype=float)
        queried_ids = self.qs.query(
            X=np.concatenate([X_l, X_u]),
            y=np.concatenate([y_l, nan_labels]),
            ensemble=learners,
            fit_ensemble=True,
            batch_size=num_queries,
        )
        queried_original_ids = queried_ids - len(y_l)
        return queried_original_ids

    def get_params(self):
        params = super().get_params()
        params["ensemble_size"] = self.ensemble_size
        return params


class QueryByCommitteeQueryStrategy(ActiveMLEnsembleQueryStrategy):
    """QueryByCommitteeQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the Query by Committee method.

    Args:
        seed (int): The seed for the random number generator.
        method (str): The method for Query by Committee.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        qs: object
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, method, ensemble_size):
        super().__init__(seed=seed, qs=QueryByCommittee(method=method, random_state=seed), ensemble_size=ensemble_size)


class EnsemblePseudoRandomizedQueryStrategy(PseudoRandomizedQueryStrategy):
    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed)
        self.ensemble_size = ensemble_size

    @abstractmethod
    def sample(self, learner, X_l, y_l, X_u, num_queries):
        pass

    def get_params(self):
        return {"seed": self.seed, "ensemble_size": self.ensemble_size}


class QueryByCommitteeEntropyQueryStrategy(QueryByCommitteeQueryStrategy):
    """QueryByCommitteeEntropyQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the Query by Committee method
    with entropy as measure of ensemble disagreement.

    Args:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed, method="vote_entropy", ensemble_size=ensemble_size)


class QueryByCommitteeKLQueryStrategy(QueryByCommitteeQueryStrategy):
    """QueryByCommitteeKLQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the Query by Committee method
    with KL-divergence as measure of ensemble disagreement.

    Args:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed, method="KL_divergence", ensemble_size=ensemble_size)


class BatchBaldQueryStrategy(ActiveMLEnsembleQueryStrategy):
    """BatchBaldQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the BatchBALD method.

    Args:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed, qs=BatchBALD(random_state=seed), ensemble_size=ensemble_size)


#########################
# self-implemented
#########################
class EmbeddingBasedQueryStrategy(PseudoRandomizedQueryStrategy):
    """EmbeddingBasedQueryStrategy

    This class is an abstract class for query strategies that are based on embeddings. The query strategies are used to
    sample instances from the pool of unlabeled instances.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    @abstractmethod
    def sample(self, learner, X_l, y_l, X_u, num_queries):
        pass

    def compute_embedding(self, learner, X_l, y_l, X_u, transform_labeled=False):
        learner_fqn = fullname(learner)
        if learner_fqn == "tabpfn.scripts.transformer_prediction_interface.TabPFNClassifier":
            from alpbench.util.TorchUtil import TabPFNEmbedder

            clf = TabPFNEmbedder(X_l, y_l)
            if not transform_labeled:
                X_u = clf.forward(X_u, encode=True)
            else:
                X = np.concatenate((X_l, X_u))
                X_embeds = clf.forward(X, encode=True)
                X_l, X_u = X_embeds[: len(X_l)], X_embeds[len(X_l) :]
        elif learner_fqn == "pytorch_tabnet.tab_model.TabNetClassifier":
            from pytorch_tabnet.tab_model import TabNetClassifier

            clf = TabNetClassifier(verbose=0)
            from alpbench.util.TorchUtil import TimeLimitCallback

            clf.fit(X_l, y_l, callbacks=[TimeLimitCallback(180)])
            if not transform_labeled:
                X_u = clf.predict_proba(X_u, get_embeds=True)
            else:
                X = np.concatenate((X_l, X_u))
                X_embeds = clf.predict_proba(X, get_embeds=True)
                X_l, X_u = X_embeds[: len(X_l)], X_embeds[len(X_l) :]
        if not transform_labeled:
            return X_u
        else:
            return X_u, X_l


class TypicalClusterQueryStrategy(EmbeddingBasedQueryStrategy):
    """TypicalClusterQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the typicality of the instances
    in the clusters.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        pool_size = len(y_l)
        num_cluster = pool_size + num_queries
        X_u, X_l = self.compute_embedding(learner, X_l=X_l, y_l=y_l, X_u=X_u, transform_labeled=True)
        kmeans = KMeans(n_clusters=num_cluster)
        X = np.concatenate((X_l, X_u))
        kmeans.fit(X)
        # We don't want labels from this group
        labeled_cluster_classes = np.unique(kmeans.labels_[:pool_size])
        # Array of the following type: position in array is the label, value is the size of the cluster
        cluster_sizes = [len(np.argwhere(kmeans.labels_ == i)) for i in range(num_cluster)]
        ids_by_size = np.argsort(-np.array(cluster_sizes))
        labels_of_sorted_clusters = np.arange(num_cluster)[ids_by_size]
        selected_ids = []
        # Create a mask for instances_ids to track selected elements
        mask = np.ones(len(kmeans.labels_), dtype=bool)
        # Iterate through the num_samples uncovered largest clusters
        for idx in ids_by_size:
            current_label = labels_of_sorted_clusters[idx]
            if current_label not in labeled_cluster_classes:
                # Get neighbors within this cluster
                instances_ids = np.argwhere(kmeans.labels_ == current_label).flatten()
                instances_ids = instances_ids[mask[instances_ids]]  # Apply the mask
                if len(instances_ids) == 0:
                    continue
                elif len(instances_ids) == 1:
                    selected_ids.append(instances_ids[0])
                    mask[instances_ids[0]] = False  # Update the mask
                    if len(selected_ids) == num_queries:
                        selected_ids = np.array(selected_ids).flatten()
                        return selected_ids - pool_size
                else:
                    instances = X[instances_ids]
                    # Compute typicality for each instance, append the one with highest typicality
                    typicalities = []
                    K = 20
                    for i, instance in enumerate(instances):
                        remaining_instances = np.delete(instances, i, axis=0)
                        dists = np.sqrt(np.sum((instance - remaining_instances) ** 2, axis=1))
                        if len(dists) < K:
                            dist = np.mean(dists)
                        else:
                            dist = np.mean(np.sort(dists)[:K])
                        typicality = 1 / (dist + 1e-8)
                        typicalities.append(typicality)
                    typicalities = np.array(typicalities)
                    selected_id = instances_ids[np.argmax(typicalities)]
                    selected_ids.append(selected_id)
                    mask[selected_id] = False  # Update the mask
                    if len(selected_ids) == num_queries:
                        selected_ids = np.array(selected_ids).flatten()
                        return selected_ids - pool_size


class LeastConfidentQueryStrategy(PseudoRandomizedQueryStrategy):
    """LeastConfidentQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on least confidence.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        probas = learner.predict_proba(X_u)
        least_confidence = 1 - np.max(probas, axis=1)
        least_confidence_ids = np.argsort(least_confidence)[-num_queries:]
        return least_confidence_ids


class EntropyQueryStrategy(PseudoRandomizedQueryStrategy):
    """EntropyQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on entropy.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        probas = learner.predict_proba(X_u)
        entropies = -np.sum(probas * np.log(probas + 1e-8), axis=1)
        entropy_ids = np.argsort(entropies)[-num_queries:]
        return entropy_ids


class MarginQueryStrategy(PseudoRandomizedQueryStrategy):
    """MarginQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the margin method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        probas = learner.predict_proba(X_u)
        sorted_probas = np.sort(probas, axis=-1)
        margins = sorted_probas[:, -1] - sorted_probas[:, -2]
        original_margin_ids = np.argsort(margins)[:num_queries]
        return original_margin_ids


class PowerMarginQueryStrategy(PseudoRandomizedQueryStrategy):
    """PowerMarginQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the power margin method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        probas = learner.predict_proba(X_u)
        sorted_probas = np.sort(probas, axis=-1)
        margins = sorted_probas[:, -1] - sorted_probas[:, -2]
        margins = 1 - np.array(margins)
        # power transform
        np.random.seed(self.seed)
        margins = np.log(margins + 1e-8) + np.random.gumbel(size=len(margins))
        original_margin_ids = np.argsort(margins)[-num_queries:]
        return original_margin_ids


class RandomMarginQueryStrategy(PseudoRandomizedQueryStrategy):
    """RandomMarginQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the random margin method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        num_for_margin = num_queries // 2
        num_for_random = num_queries - num_for_margin
        random_ids = RandomQueryStrategy(self.seed).sample(learner, X_l, y_l, X_u, num_for_random)
        if num_for_margin == 0:
            return random_ids
        else:
            margin_ids = MarginQueryStrategy(self.seed).sample(learner, X_l, y_l, X_u, num_for_margin)
            return np.concatenate((margin_ids, random_ids))


class MinMarginQueryStrategy(EnsemblePseudoRandomizedQueryStrategy):
    """MinMarginQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the minimum margin method.

    Args:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed, ensemble_size=ensemble_size)
        self.ensemble_size = ensemble_size

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        clf = Ens(estimator=learner, num_estimators=self.ensemble_size, max_neighbors=len(y_l))
        clf.fit(X_l, y_l)
        all_probas = clf.predict_proba(X_u)
        margins = np.zeros((self.ensemble_size, all_probas.shape[0]))
        sorted_probas = np.sort(all_probas, axis=1)
        margins = sorted_probas[:, -1, :] - sorted_probas[:, -2, :]
        margins = np.min(margins, axis=-1)
        margin_ids = np.argsort(margins)[:num_queries]
        return margin_ids


class FalcunQueryStrategy(PseudoRandomizedQueryStrategy):
    """FalcunQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the FALCUN method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        probas = learner.predict_proba(X_u)
        sorted_probas = np.sort(probas, axis=-1)
        margins = sorted_probas[:, -1] - sorted_probas[:, -2]
        div_scores = 1 - margins
        gamma = 10
        selected_ids = []
        ids_to_choose_from = np.arange(len(margins))
        mask = np.ones(len(margins), dtype=bool)  # Initialize mask
        for round in range(num_queries):
            relevance = margins + div_scores
            relevance[np.isnan(relevance)] = 0
            prob = relevance[mask] ** gamma / np.sum(relevance[mask] ** gamma)
            # set nan to 0, afterwards normalize
            prob[np.isnan(prob)] = 0
            prob = prob / np.sum(prob)
            prob[np.isnan(prob)] = 0
            np.random.seed(self.seed)
            # if still a nan
            if np.sum(np.isnan(prob)) > 0:
                selected_id = np.random.choice(ids_to_choose_from[mask])
            else:
                selected_id = np.random.choice(ids_to_choose_from[mask], p=prob)
            selected_ids.append(selected_id)
            # update div scores
            x_q = probas[selected_id]
            mask[selected_id] = False
            # remove selected id
            distances = np.sum(abs(probas - x_q), axis=1)
            div_scores = np.minimum(div_scores, distances)
            # normalize
            div_scores = (div_scores - np.min(div_scores)) / (np.max(div_scores) - np.min(div_scores) + 1e-8)

        return selected_ids


class WeightedClusterQueryStrategy(EmbeddingBasedQueryStrategy):
    """WeightedClusterQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the weighted cluster method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        scores = learner.predict_proba(X_u) + 1e-8
        entropy = -np.sum(scores * np.log(scores), axis=1)
        num_classes = num_queries
        X_u = self.compute_embedding(learner, X_l, y_l, X_u)

        from sklearn.cluster import KMeans
        from sklearn.metrics.pairwise import euclidean_distances

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            kmeans = KMeans(n_clusters=num_classes)
            kmeans.fit(X_u, sample_weight=entropy)
            centers = kmeans.cluster_centers_
            # for each class, find instance clostes to center
            selected_ids = []

            per_class = num_queries // num_classes + 1
            for class_ in range(num_classes):
                class_ids = np.argwhere(kmeans.labels_ == class_)
                # corresponding center
                center = centers[class_].reshape(1, -1)
                # get nearest neighbor
                class_neigh = np.squeeze(X_u[class_ids], axis=1)
                dists = euclidean_distances(center, class_neigh)
                if len(dists) < per_class:
                    closest_instances_id = np.argsort(dists)[:][0]
                else:
                    closest_instances_id = np.argsort(dists)[:per_class][0]
                for closest_instance_id in closest_instances_id:
                    selected_ids.append(class_ids[closest_instance_id][0])

            return selected_ids[0:num_queries]


class KMeansQueryStrategy(EmbeddingBasedQueryStrategy):
    """KMeansQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the KMeans method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        X_u = self.compute_embedding(learner, X_l, y_l, X_u)
        num_classes = len(np.unique(y_l))
        from sklearn.cluster import KMeans
        from sklearn.metrics.pairwise import euclidean_distances

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            kmeans = KMeans(n_clusters=num_classes)

            kmeans.fit(X_u)
            centers = kmeans.cluster_centers_
            # for each class, find instance clostes to center
            selected_ids = []

            per_class = num_queries // num_classes + 1
            for class_ in range(num_classes):
                class_ids = np.argwhere(kmeans.labels_ == class_)
                # corresponding center
                center = centers[class_].reshape(1, -1)
                class_neigh = np.squeeze(X_u[class_ids], axis=1)
                dists = euclidean_distances(center, class_neigh)
                if len(dists) < per_class:
                    closest_instances_id = np.argsort(dists)[:][0]
                else:
                    closest_instances_id = np.argsort(dists)[:per_class][0]
                for closest_instance_id in closest_instances_id:
                    selected_ids.append(class_ids[closest_instance_id][0])

            return selected_ids[0:num_queries]


class ClusterMarginQueryStrategy(EmbeddingBasedQueryStrategy):
    """ClusterMargin

    This class is used to sample instances from the pool of unlabeled instances based on the cluster margin method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        m = 10
        probas = learner.predict_proba(X_u)
        X_u = self.compute_embedding(learner, X_l, y_l, X_u)
        num_clusters = min((len(X_l) + len(X_u)) // m, len(X_u))
        clustering = AgglomerativeClustering(n_clusters=num_clusters).fit(X_u)

        sorted_probas = np.sort(probas, axis=-1)
        margins = sorted_probas[:, -1] - sorted_probas[:, -2]

        # retrieve m*num_queries instances with smallest margins
        num_to_retrieve = m * num_queries
        ids_to_retrieve = np.argsort(margins)[:num_to_retrieve]
        # from those sample diversity-based through clustering
        cluster_belongings = clustering.labels_[ids_to_retrieve]
        # sort cluster by size
        cluster_dict = {}
        for i in set(cluster_belongings):
            cluster_dict[i] = len(np.argwhere(cluster_belongings == i))
        # sort by size
        keys = np.array(list(cluster_dict.keys()))
        values = np.array(list(cluster_dict.values()))
        sorted_indices = np.argsort(values)
        sorted_keys = keys[sorted_indices]

        selected_ids = []
        mask = np.ones(len(cluster_belongings), dtype=bool)

        while len(selected_ids) < num_queries:
            for key in sorted_keys:
                to_sample_from = np.argwhere((cluster_belongings == key) & mask).flatten()
                if len(to_sample_from) == 0:
                    continue
                else:
                    np.random.seed(self.seed)
                    id = np.random.choice(to_sample_from)
                    selected_ids.append(ids_to_retrieve[id])
                    mask[id] = False
                    if len(selected_ids) == num_queries:
                        return selected_ids

        return selected_ids[:num_queries]


class MaxEntropyQueryStrategy(EnsemblePseudoRandomizedQueryStrategy):
    """MaxEntropyQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the maximum entropy method.

    Args:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed, ensemble_size=25)
        self.ensemble_size = ensemble_size

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        clf = Ens(estimator=learner, num_estimators=self.ensemble_size, max_neighbors=len(y_l))
        clf.fit(X_l, y_l)
        probas = clf.predict_proba(X_u).mean(axis=-1)
        entropies = -np.sum(probas * np.log(probas + 1e-8), axis=1)
        entropy_ids = np.argsort(entropies)[-num_queries:]
        return entropy_ids


class CoreSetQueryStrategy(EmbeddingBasedQueryStrategy):
    """CoreSetQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the core set method.

    Args:
        seed (int): The seed for the random number generator.

    Attributes:
        seed (int): The seed for the random number generator.
    """

    def __init__(self, seed):
        super().__init__(seed=seed)

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        X_u, X_l = self.compute_embedding(learner, X_l, y_l, X_u, transform_labeled=True)
        selected_ids = []
        # Initialize the mask to all True, meaning all samples are initially available
        mask = np.ones(X_u.shape[0], dtype=bool)
        for round in range(num_queries):
            active_set = X_l[:, np.newaxis, :]
            inactive_set = X_u[np.newaxis, :, :]

            # Apply the mask to inactive_set
            inactive_set_masked = inactive_set[:, mask, :]

            distances = np.linalg.norm(active_set - inactive_set_masked, axis=2)

            # compute distance to closest neighbor
            dists = distances.min(axis=0)
            # get the id of the instance with the highest distance
            selected_id = np.argmax(dists)
            original_selected_id = np.where(mask)[0][selected_id]

            selected_ids.append(original_selected_id)

            # remove selected id from the mask
            mask[original_selected_id] = False

            # add the selected instance to X_l
            X_l = np.concatenate((X_l, X_u[original_selected_id].reshape(1, -1)))

        return selected_ids


class BALDQueryStrategy(EnsemblePseudoRandomizedQueryStrategy):
    """BALDQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on the BALD method.

    Args:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed, ensemble_size=25)
        self.ensemble_size = ensemble_size

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        clf = Ens(estimator=learner, num_estimators=self.ensemble_size, max_neighbors=len(y_l))
        clf.fit(X_l, y_l)
        probas = clf.predict_proba(X_u)
        probas_mean = probas.mean(axis=-1)
        entropies = -np.sum(probas_mean * np.log(probas_mean + 1e-8), axis=1)
        mean_entropy = np.mean(
            np.array(
                [-np.sum(probas[:, :, i] * np.log(probas[:, :, i] + 1e-8), axis=1) for i in range(self.ensemble_size)]
            ),
            axis=0,
        )
        mutual_info = entropies - mean_entropy
        mutual_info_ids = np.argsort(mutual_info)[-num_queries:]
        return mutual_info_ids


class PowerBALDQueryStrategy(EnsemblePseudoRandomizedQueryStrategy):
    """PowerBALDQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on a power version of BALD.

    Args:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed, ensemble_size=25)
        self.ensemble_size = ensemble_size

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        clf = Ens(estimator=learner, num_estimators=self.ensemble_size, max_neighbors=len(y_l))
        clf.fit(X_l, y_l)
        probas = clf.predict_proba(X_u)
        probas_mean = probas.mean(axis=-1)
        entropies = -np.sum(probas_mean * np.log(probas_mean + 1e-8), axis=1)
        mean_entropy = np.mean(
            np.array(
                [-np.sum(probas[:, :, i] * np.log(probas[:, :, i] + 1e-8), axis=1) for i in range(self.ensemble_size)]
            ),
            axis=0,
        )
        mutual_info = entropies - mean_entropy + np.random.gumbel(size=len(entropies))
        mutual_info_ids = np.argsort(mutual_info)[-num_queries:]
        return mutual_info_ids


class QBCVarianceRatioQueryStrategy(EnsemblePseudoRandomizedQueryStrategy):
    """QBCVarianceRatioQueryStrategy

    This class is used to sample instances from the pool of unlabeled instances based on QBC method with variance
    ratio as measure of ensemble disagreement.

    Args:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.

    Attributes:
        seed (int): The seed for the random number generator.
        ensemble_size (int): The size of the ensemble.
    """

    def __init__(self, seed, ensemble_size):
        super().__init__(seed=seed, ensemble_size=25)
        self.ensemble_size = ensemble_size

    def sample(self, learner, X_l, y_l, X_u, num_queries):
        clf = Ens(estimator=learner, num_estimators=self.ensemble_size, max_neighbors=len(y_l))
        clf.fit(X_l, y_l)
        probas = clf.predict_proba(X_u)
        preds = probas.argmax(axis=1)
        variance_ratios = []
        for i in range(len(preds)):
            unique_labels, counts = np.unique(preds[i], return_counts=True)
            # Find the index of the label that occurs most often
            most_common_index = np.argmax(counts)
            # Get the most common label and its count
            most_common_count = counts[most_common_index]
            variance_ratio = 1 - most_common_count / self.ensemble_size
            variance_ratios.append(variance_ratio)
        variance_ratios = np.array(variance_ratios)
        vr_ids = np.argsort(variance_ratios)[-num_queries:]
        return vr_ids



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.colors import Normalize, LinearSegmentedColormap

import pandas as pd
from py_experimenter.experimenter import PyExperimenter
import json
import os
from scipy.stats import ttest_ind_from_stats
from scipy.stats import ttest_ind as tt



class BudgetPerformancePlot:
    """BudgetPerformancePlot

    This class plots the performance of different sampling strategies over the budget.

    Args:
        data (dict): The data to plot.
        path_to_save (str): The path to save the plot.

    Attributes:
        data (dict): The data to plot.
        path_to_save (str): The path to save the plot.

    """
    def __init__(self, df, openml_id, learner_name, metric, path_to_save=None):
        self.df = df
        self.openml_id = openml_id
        self.learner_name = learner_name
        self.metric = metric
        self.path_to_save = path_to_save
        self.plot_data = None

    def generate_plot_data(self):
        """
        This function generates the data to plot.
        """
        # get data for openml_id
        df = self.df[self.df['openml_id'] == self.openml_id]
        # get data for learner
        df = df[df['learner_name'] == self.learner_name]
        # get unique query strategy names
        query_strategies = df['query_strategy_name'].unique()
        # get unique budget values
        budgets = df['len_X_l'].unique()

        # create dict to store data
        data = {}
        for qs in query_strategies:
            data[qs] = {'budget': [], 'mean': [], 'std': [], 'mean_top': [], 'std_top': [], 'mean_low': [], 'std_low': [], 'accs_iter_3': {},'accs_iter_1': {},'accs_iter_2': {}, 'accs_iter_8': {}, 'accs_iter_20': {}, 'accs_iter_init': {}}
            for enum,budget in enumerate(budgets):
                # get data for query strategy and budget
                df_temp = df[(df['query_strategy_name'] == qs) & (df['len_X_l'] == budget)]

                #print("DF TEMP", df_temp)
                #return


                # reset index
                df_temp.reset_index()
                # get mean and std of metric
                mean = df_temp[self.metric].mean()
                std = df_temp[self.metric].std()
                # get ids of top 5 and low 5 values
                if enum == 0:
                    mean_top_ids = df_temp[self.metric].argsort().tolist()[-10:]
                    mean_low_ids = df_temp[self.metric].argsort().tolist()[:10]
                    data[qs]['accs_iter_init']['strong'] = df_temp[self.metric].values[mean_top_ids]
                    data[qs]['accs_iter_init']['weak'] = df_temp[self.metric].values[mean_low_ids]
                if enum == 1:
                    data[qs]['accs_iter_1']['strong'] = df_temp[self.metric].values[mean_top_ids]
                    data[qs]['accs_iter_1']['weak'] = df_temp[self.metric].values[mean_low_ids]
                if enum == 2:
                    data[qs]['accs_iter_2']['strong'] = df_temp[self.metric].values[mean_top_ids]
                    data[qs]['accs_iter_2']['weak'] = df_temp[self.metric].values[mean_low_ids]
                if enum == 3:
                    data[qs]['accs_iter_3']['strong'] = df_temp[self.metric].values[mean_top_ids]
                    data[qs]['accs_iter_3']['weak'] = df_temp[self.metric].values[mean_low_ids]
                if enum == 8:
                    data[qs]['accs_iter_8']['strong'] = df_temp[self.metric].values[mean_top_ids]
                    data[qs]['accs_iter_8']['weak'] = df_temp[self.metric].values[mean_low_ids]
                if enum == 20:
                    data[qs]['accs_iter_20']['strong'] = df_temp[self.metric].values[mean_top_ids]
                    data[qs]['accs_iter_20']['weak'] = df_temp[self.metric].values[mean_low_ids]

                df_reset = df_temp[self.metric].reset_index()
                mean_top = np.array(df_reset.loc[mean_top_ids][self.metric]).mean()
                mean_low = np.array(df_reset.loc[mean_low_ids][self.metric]).mean()
                std_top = np.array(df_reset.loc[mean_top_ids][self.metric]).std()
                std_low = np.array(df_reset.loc[mean_low_ids][self.metric]).std()

                # append data to dict
                data[qs]['budget'].append(budget)
                data[qs]['mean'].append(mean)
                data[qs]['std'].append(std)
                data[qs]['mean_top'].append(mean_top)
                data[qs]['std_top'].append(std_top)
                data[qs]['mean_low'].append(mean_low)
                data[qs]['std_low'].append(std_low)


        self.plot_data = data


    def show(self):
        """
        This function plots the performance of different query strategies over the budget and saves
        it as .pdf under the specified path.
        """
        data = self.plot_data
        if len(data.keys()) == 0:
            return
        else:
            oid = self.openml_id
            # we want to order the QS by grouping them into categories, also each category gets a different
            # color coding (uncertainty based are redish, representative are greenish, hybrid are blueish,
            # random is pink)
            keys = data.keys()
            list_of_qs = list(keys)
            all_qs_ordered = ["random", "entropy", "least_confident", "margin", "power_margin",
                        "max_entropy", "bald", "power_bald", "qbc_variance_ratio",
                        "kmeans", "core_set","typ_cluster",
                        "cluster_margin", "falcun",  "weighted_cluster"]
            # order the query strategies
            ordered_qs = []
            for qs in all_qs_ordered:
                if qs in list_of_qs:
                    ordered_qs.append(qs)

            fig, ax = plt.subplots(1)
            # colormap of len(keys)
            colors = plt.cm.tab20(np.linspace(0, 1, len(keys)))
            color_dict = {"random": "magenta", "least_confident": "rosybrown", "margin": "red",
                          "entropy": "orange", "power_margin": "brown", "bald": "tomato",
                          "power_bald": "coral", "max_entropy": "sandybrown", "qbc_variance_ratio": "peachpuff",
                          "core_set": "limegreen", "typ_cluster": "forestgreen", "cluster_margin": "mediumblue",
                          "weighted_cluster": "turquoise", "falcun": "blue"}
            for cl, key in enumerate(ordered_qs):
                budget = np.array(data[key]['budget'])
                mu = np.array(data[key]['mean'])
                std = np.array(data[key]['std'])/np.sqrt(10)
                mu_top = np.array(data[key]['mean_top'])
                std_top = np.array(data[key]['std_top'])/np.sqrt(10)
                mu_low = np.array(data[key]['mean_low'])
                std_low = np.array(data[key]['std_low'])/np.sqrt(10)
                qs = key
                cl = color_dict[qs]
                #plt.plot(budget, mu, lw=2, label=key, color=cl)
                #plt.fill_between(budget, mu+std, mu-std,  facecolor=cl, alpha=0.5)
                #plt.plot(budget, mu_top, lw=2, label=key, color=cl)
                #plt.plot(budget, mu_low, lw=2,  color=cl, linestyle='--')
                plt.fill_between(budget, mu_top+std_top, mu_top-std_top,  facecolor=cl, alpha=0.5)
                plt.fill_between(budget, mu_low+std_low, mu_low-std_low,  facecolor=cl, alpha=0.5)
            for cl, key in enumerate(ordered_qs):
                budget = np.array(data[key]['budget'])
                mu = np.array(data[key]['mean'])
                std = np.array(data[key]['std'])/np.sqrt(10)
                mu_top = np.array(data[key]['mean_top'])
                std_top = np.array(data[key]['std_top'])/np.sqrt(10)
                mu_low = np.array(data[key]['mean_low'])
                std_low = np.array(data[key]['std_low'])/np.sqrt(10)
                qs = key
                cl = color_dict[qs]
                #plt.plot(budget, mu, lw=2, label=key, color=cl)
                #plt.fill_between(budget, mu+std, mu-std,  facecolor=cl, alpha=0.5)
                plt.plot(budget, mu_top, lw=2, label=key, color=cl)
                plt.plot(budget, mu_low, lw=2,  color=cl, linestyle='--')
                #plt.fill_between(budget, mu_top+std_top, mu_top-std_top,  facecolor=cl, alpha=0.5)
                #plt.fill_between(budget, mu_low+std_low, mu_low-std_low,  facecolor=cl, alpha=0.5)

            qs_for_figure = {"margin": "MS", "least_confident": "LC", "entropy": "ES",
                                "power_margin": "PowMS", "bald": "BALD", "power_bald": "PowBALD",
                                "max_entropy": "MaxEnt", "qbc_variance_ratio": "QBC VR",
                                "core_set": "CoreSet", "typ_cluster": "TypClu", "cluster_margin": "CluMS",
                                "weighted_cluster": "Clue", "falcun": "FALCUN", "random": "Rand"}
            learner_for_figure = {"knn_3": "KNN", "svm_rbf": "SVM", "mlp": "MLP", "rf_entropy": "RF",
                                    "catboost": "Catboost", "xgb": "XGB", "tabnet": "Tabnet", "tabpfn": "TabPFN"}

            handles, labels = plt.gca().get_legend_handles_labels()

            new_labels = []
            for label in labels:
                new_labels.append(qs_for_figure[label])

            plt.title(learner_for_figure[self.learner_name] + " on id " + str(self.openml_id), fontsize=30)
            plt.legend(handles, new_labels, fontsize=20, loc='lower right')# fontsize=25,loc='center left', bbox_to_anchor=(1, 0.5))
            plt.yticks(fontsize=20)
            plt.xticks(fontsize=20)
            ax.set_xlabel('Number of labeled instances', fontsize=25)
            ax.set_ylabel('test accuracy', fontsize=25)
            [l.set_visible(False) for (i, l) in enumerate(ax.xaxis.get_ticklabels()) if i % 2 != 0]

            # save image
            if self.path_to_save is not None:
                if not os.path.exists(self.path_to_save):
                    os.makedirs(self.path_to_save)
            else:
                self.path_to_save = "FIGURES/BUDGET_PERFORMANCE_PLOT/" +  str(self.openml_id) + "/"
                if not os.path.exists(self.path_to_save):
                    os.makedirs(self.path_to_save)
            self.path_to_save = self.path_to_save + str(self.learner_name) + ".pdf"
            fig.savefig(self.path_to_save, facecolor='white', transparent=True, bbox_inches='tight')
            plt.close()



class HeatMapPlot:
    """HeatMapPlot

    This class plots a heatmap of the performance of different active learning pipelines as well as win or lose-
    matrices for the specified learner comparing different query strategies.

    Args:
        data (dict): The data to plot.
        path_to_save (str): The path to save the plot.
        filter_ids (str): The filter ids.
        take_statistical_insignificant (bool): Whether to take statistical insignificant values.

    Attributes:
        data (dict): The data to plot.
        path_to_save (str): The path to save the plot.
        filter_ids (str): The filter ids.
        take_statistical_insignificant (bool): Whether to take statistical insignificant values.

    """
    def __init__(self, df, learner_name, path_to_save = None, filter_ids = "all", take_statistical_insignificant=False):
        self.df = df
        self.path_to_save = path_to_save
        self.filter_ids = filter_ids
        self.take_statistical_insignificant = take_statistical_insignificant


    def show_heatmap(self):
        """
        This function plots heatmaps to compare performances of different active learning pipelines.
        The figures are saved under the specified path.
        """
        learners_ordered = ["knn_3", "svm_rbf", "rf_entropy", "catboost",
                            "xgb","mlp",
                            "tabnet",
                            "tabpfn"]
        info = ["margin", "entropy", "power_margin",  "power_bald"]
        repr = ["core_set", "typ_cluster"]
        hybr = ["cluster_margin", "falcun"]

        sampling_strategies_ordered = ["random",
                                       "margin",
                                       # "least_confident",
                                       "entropy", #"qbc_variance_ratio",
                                       #"max_entropy",
                                       "power_margin", #"bald",
                                       "power_bald",
                                       "core_set", #"kmeans",
                                       "typ_cluster",
                                       "cluster_margin",
                                       #"weighted_cluster",
                                       "falcun"
                                       ]



        learners = self.data['learners']
        sampling_strategies = self.data['sampling_strategies']
        heatmap = self.data['heatmap']


        res = np.zeros((len(learners_ordered), len(sampling_strategies_ordered)))
        for enum_i, l in enumerate(learners_ordered):
            for enum_j, qs in enumerate(sampling_strategies_ordered):
                res[enum_i, enum_j] = heatmap[(l,qs)]


        setting_name = self.data['setting_name']
        metric_name = self.data['metric_name'] + "_AUBC"


        # Define a custom colormap from light red to red
        light_red_to_red = LinearSegmentedColormap.from_list('light_red_to_red', ['lavender', 'darkblue'])
        greens = plt.cm.Greens
        reds = plt.cm.Reds
        blues = plt.cm.Blues
        purples = plt.cm.Purples

        # Normalize the data
        norm = Normalize(vmin=res.min(), vmax=res.max())

        # Apply the custom colormap to the data
        colors = light_red_to_red(norm(res))
        red_colors = reds(norm(res))
        green_colors = greens(norm(res))
        blue_colors = blues(norm(res))
        purple_colors = purples(norm(res))



        # Erstellen des Plots
        fig, ax = plt.subplots(figsize=(1.5 * len(sampling_strategies_ordered), 1.5 * len(learners_ordered)))

        # Zeichnen der Tafel mit den Farben basierend auf den Werten
        for i,l in enumerate(learners_ordered):
            for j,qs in enumerate(sampling_strategies_ordered):
                #print("hey", str(int(res_wins[i,j])) + "/" + str(int(res_losses[i,j])))
                #print("total " , int(res_wins[i,j] + res_losses[i,j]))
                ax.text(j + 0.5, i + 0.5, str(int(res[i,j])), ha='center', va='center', color='white',
                        fontsize=30,
                        weight='bold')
                if qs in info:
                    colors = red_colors
                elif qs in repr:
                    colors = green_colors
                elif qs in hybr:
                    colors = blue_colors
                else:
                    colors = purple_colors
                ax.add_patch(plt.Rectangle((j, i), 1, 1, fill=True, color=colors[i, j]))

        # Einstellungen für den Plot
        x_pos = np.arange(1, len(sampling_strategies_ordered) + 1.5, 1) - .5
        y_pos = np.arange(1, len(learners_ordered) + 1.5, 1) - .5


        x_pos[-1] -= .5
        y_pos[-1] -= .5

        ax.set_xticks(x_pos)
        ax.set_yticks(y_pos)

        learners_ordered = ["KNN", "SVM",  "RF", "Catboost", "XGB","MLP",
                            "Tabnet",
                            "TabPFN"]
        sampling_strategies_ordered = ["Rand",
                                       "MS", #"least_confident",
                                       "ES",  # "qbc_variance_ratio",
                                       # "max_entropy",
                                       "PowMS",  # "bald",
                                       "PowBALD",
                                       "CoreSet",
                                       "TypClu",
                                       # "kmeans",
                                       "CluMS",
                                       # "weighted_cluster",
                                       "FALCUN"
                                       ]

        sampling_names = list(sampling_strategies_ordered)
        learners_names = list(learners_ordered)

        sampling_names.append(" ")
        learners_names.append(" ")
        ax.set_xticklabels(sampling_names, fontsize=40, rotation=45)
        ax.set_yticklabels(learners_names, fontsize=40)
        if self.take_statistical_insignificant:
            fig.suptitle("Setting: " + setting_name + ", Datasets: " + str(self.filter_ids),
                     fontsize=50)
            ax.set_title("(statistically significant)", fontsize=30)

        else:
            fig.suptitle("Setting: " + setting_name + ", Datasets: " + str(self.filter_ids),
                     fontsize=50)
            ax.set_title("(not statistically significant)", fontsize=30)#, color='white')

        filter_ids = self.data['filter_ids']
        if self.path_to_save is not None:
            PATH = self.path_to_save + str(filter_ids) +"/"
            if not os.path.exists(PATH):
                os.makedirs(PATH)
            SAVE_PATH = PATH + setting_name +  metric_name +  "_"+str(filter_ids) +"_.pdf"
        print("saved in ", SAVE_PATH)
        fig.savefig(SAVE_PATH, facecolor='white', transparent=True, bbox_inches='tight')
        #plt.show()



class WinMatrixPlot:
    def __init__(self, df, learner_name, path_to_save = None, filter_ids = "all", take_statistical_insignificant=False):
        self.df = df
        self.learner_name = learner_name
        self.path_to_save = path_to_save
        self.filter_ids = filter_ids
        self.take_statistical_insignificant = take_statistical_insignificant

    def generate_win_matrix(self):
        """
        This function generates win-matrices to compare performances of different query strategies combined
        with one fixed learning algorithm.
        """
        df = self.df[self.df['learner_name'] == self.learner_name]
        win_matrix = {}
        query_strategies = self.df['query_strategy_name'].unique()
        for qs1 in query_strategies:
            for qs2 in query_strategies:
                win_matrix[(qs1, qs2)] = [0, 0]

        for oid in df['openml_id'].unique():
            df = df[df['learner_name'] == self.learner_name]
            for qs1 in query_strategies:
                for qs2 in query_strategies:
                    df1 = df[df['query_strategy_name'] == qs1]
                    df2 = df[df['query_strategy_name'] == qs2]
                    mean1 = df1['test_accuracy'].mean()
                    mean2 = df2['test_accuracy'].mean()
                    std1 = df1['test_accuracy'].std()
                    std2 = df2['test_accuracy'].std()
                    t, p = ttest_ind_from_stats(mean1, std1, len(df1), mean2, std2, len(df2))
                    if p < 0.05:
                        if mean1 > mean2:
                            win_matrix[(qs1, qs2)][0] += 1
                        else:
                            win_matrix[(qs1, qs2)][1] += 1





    def show(self):
        """
        This function generates win-matrices to compare performances of different query strategies combined
        with one fixed learning algorithm.
        """

        all_qs_ordered = ["random", "entropy", "least_confident", "margin", "power_margin",
                          "max_entropy", "bald", "power_bald", "qbc_variance_ratio",
                          "kmeans", "core_set", "typ_cluster",
                          "cluster_margin", "falcun", "weighted_cluster"]

        learners_ordered = ["knn_3", "svm_rbf", "mlp","rf_entropy","catboost", "xgb",
        "tabnet", "tabpfn"]

        filter_ids = self.filter_ids




        win_matrix = self.data['learner'][learner]
        print("WIN MATRIX",win_matrix)
        sampling_strategies = win_matrix.keys()


        # create numpy array from dict
        res_wins = np.zeros((len(sampling_strategies_ordered), len(sampling_strategies_ordered)))
        res_losses  = np.zeros((len(sampling_strategies_ordered), len(sampling_strategies_ordered)))

        for i, strat_1 in enumerate(sampling_strategies_ordered):
            for j, strat_2 in enumerate(sampling_strategies_ordered):
                key = ((learner, strat_1), (learner, strat_2))

                # check if key exists in win_matrix
                if key in win_matrix.keys():
                    wins = win_matrix[key][0]
                    losses = win_matrix[key][1]
                else:
                    wins = 0
                    losses = 0
                if i == j:
                    wins = 0
                    losses = 0

                res_wins[i, j] = wins
                res_losses[i, j] = losses


        res_total = np.ones((len(sampling_strategies_ordered), len(sampling_strategies_ordered))) * self.data['num_dataset_ids']
        # set diag to 0
        np.fill_diagonal(res_total, 0)


        setting_name = self.data['setting_name']
        metric_name =  self.data['metric_name'] + "_AUBC"


        light_red_to_red = LinearSegmentedColormap.from_list('light_red_to_red', ['lavender', 'darkblue'])

        greens = plt.cm.Greens
        reds = plt.cm.Reds
        blues = plt.cm.Blues
        purples = plt.cm.Purples


        # Normalize the data
        norm = Normalize(vmin=res_wins.min(), vmax=res_wins.max())

        # Apply the custom colormap to the data
        colors = light_red_to_red(norm(res_wins))

        # Apply the custom colormap to the data
        colors = light_red_to_red(norm(res_wins))
        red_colors = reds(norm(res_wins))
        green_colors = greens(norm(res_wins))
        blue_colors = blues(norm(res_wins))
        purple_colors = purples(norm(res_wins))

        #norm = Normalize(res_wins.min(), vmax=res_wins.max() * 1.5)
        #colors = plt.cm.viridis(norm(res_wins))
        #colors = plt.cm.Reds(norm(res_wins))
        #print("COOLS",colors)
        # Use the 'Reds' colormap for red tones
        #colors = plt.cm.Reds(norm(res_wins))

        # Erstellen des Plots
        fig, ax = plt.subplots(figsize=(1.5 * len(sampling_strategies_ordered), 1.5 * len(sampling_strategies_ordered)))

        # Zeichnen der Tafel mit den Farben basierend auf den Werten
        for i,l in enumerate(sampling_strategies_ordered):
            for j,qs in enumerate(sampling_strategies_ordered):
                #print("hey", str(int(res_wins[i,j])) + "/" + str(int(res_losses[i,j])))
                #print("total " , int(res_wins[i,j] + res_losses[i,j]))

                ax.text(j + 0.5, i + 0.5,
                        str(int(res_wins[i, j])) + "/" + str(int(res_total[i, j])),
                        ha='center', va='center', color='white',
                        fontsize=25,
                        weight='bold')
                if l in info:
                    colors = red_colors
                elif l in repr:
                    colors = green_colors
                elif l in hybr:
                    colors = blue_colors
                else:
                    colors = purple_colors

                ax.add_patch(plt.Rectangle((j, i), 1, 1, fill=True, color=colors[i, j]))

        # Einstellungen für den Plot
        x_pos = np.arange(1, len(sampling_strategies_ordered) + 1.5, 1) - .5
        y_pos = np.arange(1, len(sampling_strategies_ordered) + 1.5, 1) - .5


        x_pos[-1] -= .5
        y_pos[-1] -= .5
        print("x_pos", x_pos)
        print("y_pos", y_pos)

        ax.set_xticks(x_pos)
        ax.set_yticks(y_pos)

        sampling_names = ["Rand", "MS", "ES", "PowMS", "PowBALD", "CoreSet", "TypClu", "CluMS", "FALCUN"]
                          # "LC",
                          #"ES",  #"QBC",
                                      # "MaxEnt",
                          #"PowMS", #"BALD",
                          #"PowBALD",
                          # "CoreSet",
                          #"KMeans",
                          #"CluMS",
                          #"TypClu"#,"CLUE",
                          #"FALCUN"
                          #]
        #sampling_names = list(sampling_strategies_ordered)
        learners_names = list(learners)

        sampling_names.append(" ")

        learners_ordered = ["knn_3", "svm_rbf", "mlp", "rf_entropy", "catboost", "xgb",
                            "tabnet", "tabpfn"]

        if learner == "knn_3":
            learner_name = "KNN"
        elif learner == "svm_rbf":
            learner_name = "SVM"
        elif learner == "mlp":
            learner_name = "MLP"
        elif learner == "rf_entropy":
            learner_name = "RF"
        elif learner == "catboost":
            learner_name = "Catboost"
        elif learner == "xgb":
            learner_name = "XGB"
        elif learner == "tabnet":
            learner_name = "Tabnet"
        elif learner == "tabpfn":
            learner_name = "TabPFN"


        learners_names.append(" ")
        ax.set_xticklabels(sampling_names, fontsize=40, rotation=45)
        ax.set_yticklabels(sampling_names, fontsize=40)
        ax.set_title("Setting: " + setting_name + ", Learner: " + learner_name,
                     fontsize=50)
        SAVE_PATH = "fig .pdf"


        if self.path_to_save is not None:
            PATH = self.path_to_save + "/" + str(filter_ids) +"/"
            if not os.path.exists(PATH):
                os.makedirs(PATH)
            SAVE_PATH = PATH + setting_name +  metric_name + "_"+str(learner) + ".pdf"
            print("SAVE PATH", SAVE_PATH)
            fig.savefig(SAVE_PATH, facecolor='white', transparent=True, bbox_inches='tight')
        #plt.show()






class WinMatrixPlot:
    """WinMatrixPlot

    This class plots the win-matrix to compare performances of different query strategies combined with one fixed
    learning algorithm.

    Args:
        df (pd.DataFrame): The dataframe with the computed final AUBC per pipeline.
        learner_name (str): The fixed learner name to generate the win-matrix for.
        path_to_save (str): The path to save the plot.
        take_statistical_insignificant (bool): If True, statistical insignificant results are also considered.

    Attributes:
        df (pd.DataFrame): The dataframe with the computed final AUBC per pipeline.
        learner_name (str): The fixed learner name to generate the win-matrix for.
        path_to_save (str): The path to save the plot.
        take_statistical_insignificant (bool): If True, statistical insignificant results are also considered.
        win_matrix (dict): The win-matrix to compare performances of different query strategies combined with one fixed
        learning algorithm.

    """
    def __init__(self, df, learner_name, path_to_save = None, take_statistical_insignificant=False):
        self.df = df
        self.learner_name = learner_name
        self.path_to_save = path_to_save
        self.take_statistical_insignificant = take_statistical_insignificant
        self.win_matrix = None

    def generate_win_matrix(self):
        """
        This function generates win-matrices to compare performances of different query strategies combined
        with one fixed learning algorithm.
        """
        df = self.df[self.df['learner_name'] == self.learner_name]
        win_matrix = {}
        query_strategies = self.df['query_strategy_name'].unique()
        for qs1 in query_strategies:
            for qs2 in query_strategies:
                win_matrix[(qs1, qs2)] = [0, 0]

        for oid in df['openml_id'].unique():
            df = df[df['learner_name'] == self.learner_name]
            for qs1 in query_strategies:
                for qs2 in query_strategies:
                    df1 = df[df['query_strategy_name'] == qs1]
                    df2 = df[df['query_strategy_name'] == qs2]
                    mean1 = df1['aubc'].mean()
                    mean2 = df2['aubc'].mean()
                    std1 = df1['aubc'].std()
                    std2 = df2['aubc'].std()
                    t, p = ttest_ind_from_stats(mean1, std1, len(df1), mean2, std2, len(df2))
                    if not self.take_statistical_insignificant and p < 0.05:
                        if mean1 > mean2:
                            win_matrix[(qs1, qs2)][0] += 1
                        else:
                            win_matrix[(qs1, qs2)][1] += 1
                    if self.take_statistical_insignificant:
                        if mean1 > mean2:
                            win_matrix[(qs1, qs2)][0] += 1
                        else:
                            win_matrix[(qs1, qs2)][1] += 1
            self.win_matrix = win_matrix
            return win_matrix


    def show(self):
        """
        This function generates win-matrices to compare performances of different query strategies combined
        with one fixed learning algorithm.
        """

        all_qs_ordered = ["random", "entropy", "least_confident", "margin", "power_margin",
                          "max_entropy", "bald", "power_bald", "qbc_variance_ratio",
                          "kmeans", "core_set", "typ_cluster",
                          "cluster_margin", "falcun", "weighted_cluster"]

        learners_ordered = ["knn_3", "svm_rbf", "mlp","rf_entropy","catboost", "xgb",
        "tabnet", "tabpfn"]

        # TODO implement this function bla

        query_strategies = self.win_matrix.keys()


        # create numpy array from dict
        res_wins = np.zeros((len(sampling_strategies_ordered), len(sampling_strategies_ordered)))
        res_losses  = np.zeros((len(sampling_strategies_ordered), len(sampling_strategies_ordered)))

        for i, strat_1 in enumerate(sampling_strategies_ordered):
            for j, strat_2 in enumerate(sampling_strategies_ordered):
                key = ((learner, strat_1), (learner, strat_2))

                # check if key exists in win_matrix
                if key in win_matrix.keys():
                    wins = win_matrix[key][0]
                    losses = win_matrix[key][1]
                else:
                    wins = 0
                    losses = 0
                if i == j:
                    wins = 0
                    losses = 0

                res_wins[i, j] = wins
                res_losses[i, j] = losses


        res_total = np.ones((len(sampling_strategies_ordered), len(sampling_strategies_ordered))) * self.data['num_dataset_ids']
        # set diag to 0
        np.fill_diagonal(res_total, 0)


        setting_name = self.data['setting_name']
        metric_name =  self.data['metric_name'] + "_AUBC"


        light_red_to_red = LinearSegmentedColormap.from_list('light_red_to_red', ['lavender', 'darkblue'])

        greens = plt.cm.Greens
        reds = plt.cm.Reds
        blues = plt.cm.Blues
        purples = plt.cm.Purples


        # Normalize the data
        norm = Normalize(vmin=res_wins.min(), vmax=res_wins.max())

        # Apply the custom colormap to the data
        colors = light_red_to_red(norm(res_wins))

        # Apply the custom colormap to the data
        colors = light_red_to_red(norm(res_wins))
        red_colors = reds(norm(res_wins))
        green_colors = greens(norm(res_wins))
        blue_colors = blues(norm(res_wins))
        purple_colors = purples(norm(res_wins))

        #norm = Normalize(res_wins.min(), vmax=res_wins.max() * 1.5)
        #colors = plt.cm.viridis(norm(res_wins))
        #colors = plt.cm.Reds(norm(res_wins))
        #print("COOLS",colors)
        # Use the 'Reds' colormap for red tones
        #colors = plt.cm.Reds(norm(res_wins))

        # Erstellen des Plots
        fig, ax = plt.subplots(figsize=(1.5 * len(sampling_strategies_ordered), 1.5 * len(sampling_strategies_ordered)))

        # Zeichnen der Tafel mit den Farben basierend auf den Werten
        for i,l in enumerate(sampling_strategies_ordered):
            for j,qs in enumerate(sampling_strategies_ordered):
                #print("hey", str(int(res_wins[i,j])) + "/" + str(int(res_losses[i,j])))
                #print("total " , int(res_wins[i,j] + res_losses[i,j]))

                ax.text(j + 0.5, i + 0.5,
                        str(int(res_wins[i, j])) + "/" + str(int(res_total[i, j])),
                        ha='center', va='center', color='white',
                        fontsize=25,
                        weight='bold')
                if l in info:
                    colors = red_colors
                elif l in repr:
                    colors = green_colors
                elif l in hybr:
                    colors = blue_colors
                else:
                    colors = purple_colors

                ax.add_patch(plt.Rectangle((j, i), 1, 1, fill=True, color=colors[i, j]))

        # Einstellungen für den Plot
        x_pos = np.arange(1, len(sampling_strategies_ordered) + 1.5, 1) - .5
        y_pos = np.arange(1, len(sampling_strategies_ordered) + 1.5, 1) - .5


        x_pos[-1] -= .5
        y_pos[-1] -= .5
        print("x_pos", x_pos)
        print("y_pos", y_pos)

        ax.set_xticks(x_pos)
        ax.set_yticks(y_pos)

        sampling_names = ["Rand", "MS", "ES", "PowMS", "PowBALD", "CoreSet", "TypClu", "CluMS", "FALCUN"]
                          # "LC",
                          #"ES",  #"QBC",
                                      # "MaxEnt",
                          #"PowMS", #"BALD",
                          #"PowBALD",
                          # "CoreSet",
                          #"KMeans",
                          #"CluMS",
                          #"TypClu"#,"CLUE",
                          #"FALCUN"
                          #]
        #sampling_names = list(sampling_strategies_ordered)
        learners_names = list(learners)

        sampling_names.append(" ")

        learners_ordered = ["knn_3", "svm_rbf", "mlp", "rf_entropy", "catboost", "xgb",
                            "tabnet", "tabpfn"]

        if learner == "knn_3":
            learner_name = "KNN"
        elif learner == "svm_rbf":
            learner_name = "SVM"
        elif learner == "mlp":
            learner_name = "MLP"
        elif learner == "rf_entropy":
            learner_name = "RF"
        elif learner == "catboost":
            learner_name = "Catboost"
        elif learner == "xgb":
            learner_name = "XGB"
        elif learner == "tabnet":
            learner_name = "Tabnet"
        elif learner == "tabpfn":
            learner_name = "TabPFN"


        learners_names.append(" ")
        ax.set_xticklabels(sampling_names, fontsize=40, rotation=45)
        ax.set_yticklabels(sampling_names, fontsize=40)
        ax.set_title("Setting: " + setting_name + ", Learner: " + learner_name,
                     fontsize=50)
        SAVE_PATH = "fig .pdf"


        if self.path_to_save is not None:
            PATH = self.path_to_save + "/" + str(filter_ids) +"/"
            if not os.path.exists(PATH):
                os.makedirs(PATH)
            SAVE_PATH = PATH + setting_name +  metric_name + "_"+str(learner) + ".pdf"
            print("SAVE PATH", SAVE_PATH)
            fig.savefig(SAVE_PATH, facecolor='white', transparent=True, bbox_inches='tight')
        #plt.show()



def format_insert_query(table_name, values: dict):
    """
    Format an insert query for a given table and values.

    Parameters:
        table_name (str): The name of the table to insert into.
        values (dict): A dictionary of column names and values to insert.

    Returns:
        str: The formatted insert query.
    """
    key_list = list(values.keys())

    query = f"INSERT INTO {table_name}"
    query = query + " (" + ", ".join(key_list) + ")"
    query = query + " VALUES ('" + "', '".join([str(values[key]) for key in key_list]) + "');"
    return query


def format_select_query(table_name, where: dict = None):
    """
    Format a select query for a given table and where clause.

    Parameters:
        table_name (str): The name of the table to select from.
        where (dict): A dictionary of column names and values to filter by.

    Returns:
        str: The formatted select query.
    """
    query = f"SELECT * FROM {table_name}"
    if where is not None and len(where) > 0:
        query += " WHERE " + " AND ".join([key + "='" + str(value) + "'" for key, value in where.items()])
    query = query + ";"
    return query


def instantiate_class_by_fqn(learner_fqn, learner_params={}):
    """
    Instantiate a class by its fully qualified name.

    Parameters:
        learner_fqn (str): The fully qualified name of the class to instantiate.
        learner_params (dict): A dictionary of parameters to pass to the class constructor

    Returns:
        object: The instantiated class.
    """
    import importlib
    module_name, class_name = learner_fqn.rsplit(".", 1)
    LearnerClass = getattr(importlib.import_module(module_name), class_name)
    instance = LearnerClass(**learner_params)
    return instance


def fullname(o):
    """
    Get the fully qualified name of a class.

    Parameters:
        o (object): The object to get the fully qualified name of.

    Returns:
        str: The fully qualified name of the class.
    """
    klass = o.__class__
    module = klass.__module__
    if module == "__builtin__":
        return klass.__name__
    return module + "." + klass.__name__



import numpy as np
import openml
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, OrdinalEncoder

from alpbench.benchmark.ActiveLearningSetting import ActiveLearningSetting


def create_dataset_split(
    X, y, test_split_seed, test_split_size: float, train_split_seed, train_split_size, train_split_type, factor
):
    """This method creates a split of the data into labeled, unlabeled and test data. The type of the split can be
    either absolute (i.e., a fixed number of labeled data points) or relative (i.e., a fixed share of the training
    data). The split is stratified according to the labels. The labeled data is guaranteed to contain at least one
    instance of each class. Further, if a factor is given, the number of labeled data points is determined by the
    number of classes times the factor.

    Args:
        X (numpy.ndarray): data
        y (numpy.ndarray): labels
        test_split_seed (int): seed for the test split
        test_split_size (float): size of the test data
        train_split_seed (int): seed for the train split
        train_split_size (float): size of the labeled training data
        train_split_type (str): type of the size parameter: number of data points or share of the (training) dataset
        factor (int): task-dependent factor

    Returns:
        labeled_indices (list): indices of the labeled data
        test_indices (list): indices of the test data
    """

    # initialize list of indices
    indices = np.arange(0, len(X))

    # split data into train and test and retrieve test_indices to be returned later
    X_train, X_test, y_train, y_test, train_indices, test_indices = train_test_split(
        X, y, indices, test_size=test_split_size, random_state=test_split_seed, stratify=y
    )
    # determine the proportion of unlabeled data, also in case the train split is given in terms of an absolute number
    # of labeled data points
    unlabeled_size = 1 - train_split_size
    if train_split_type == "absolute":
        if factor != -1:
            train_split_size = factor * len(np.unique(y))
            unlabeled_size = 1 - train_split_size / len(X_train)
        else:
            unlabeled_size = 1 - train_split_size / len(X_train)

    # split data into labeled and unlabeled
    X_l, X_u, y_l, y_u, labeled_indices, unlabeled_indices = train_test_split(
        X_train, y_train, train_indices, test_size=unlabeled_size, random_state=train_split_seed, stratify=y_train
    )

    if len(np.unique(y[labeled_indices])) != len(np.unique(y)):
        # make sure that each class within y is at least once in the labeled data
        for i in np.unique(y):
            if i not in y_l:
                ids = np.where(y_u == i)[0]
                np.random.seed(train_split_seed)
                idx_in_yu = np.random.choice(ids)
                idx = unlabeled_indices[idx_in_yu]
                labeled_indices = np.append(labeled_indices, idx)

    assert len(np.unique(y[labeled_indices])) == len(
        np.unique(y)
    ), "Not all classes are represented in the labeled data"

    return labeled_indices.tolist(), test_indices.tolist()


class ActiveLearningScenario:
    """Active Learning Scenario

    The active learning scenario defines the data and the setting of one active learning setup.  The scenario is
    initialized with the openml id of the dataset, the test split, train split and the seed for reproducibility, the
    setting, and optionally labeled and test indices.

    Args:
        scenario_id (int): id of the scenario in the database
        openml_id (int): id of the dataset on openml
        test_split_seed (int): seed for the test split
        train_split_seed (int): seed for the train split
        seed (int): seed for reproducibility
        setting (ActiveLearningSetting): active learning setting
        labeled_indices (list): indices of the labeled data
        test_indices (list): indices of the test data

    Attributes:
        scenario_id (int): id of the scenario in the database
        openml_id (int): id of the dataset on openml
        test_split_seed (int): seed for the test split
        train_split_seed (int): seed for the train split
        seed (int): seed for reproducibility
        setting (ActiveLearningSetting): active learning setting
        labeled_indices (list): indices of the labeled data
        test_indices (list): indices of the test data

    """

    def __init__(
        self,
        scenario_id,
        openml_id,
        test_split_seed,
        train_split_seed,
        seed,
        setting: ActiveLearningSetting,
        labeled_indices: list = None,
        test_indices: list = None,
    ):
        self.scenario_id = scenario_id
        self.openml_id = openml_id
        self.test_split_seed = test_split_seed
        self.train_split_seed = train_split_seed
        self.seed = seed
        self.labeled_indices = labeled_indices
        self.test_indices = test_indices
        self.setting = setting

        # actual data
        ds = openml.datasets.get_dataset(
            openml_id, download_data=True, download_qualities=True, download_features_meta_data=True
        )
        df = ds.get_data()[0]
        # prepare label column as numpy array
        X = np.array(df.drop(columns=[ds.default_target_attribute]).values)
        y = np.array(df[ds.default_target_attribute].values)
        if y.dtype != int:
            y_int = np.zeros(len(y)).astype(int)
            vals = np.unique(y)
            for i, val in enumerate(vals):
                mask = y == val
                y_int[mask] = i
            y = y_int
        X = OrdinalEncoder().fit_transform(X)
        X = SimpleImputer(missing_values=np.nan, strategy="mean").fit_transform(X)

        # filter X for duplicates
        _, unique_indices = np.unique(X, axis=0, return_index=True)

        self.X = X[unique_indices]
        self.y = LabelEncoder().fit_transform(y)[unique_indices]

        if test_indices is None or labeled_indices is None:
            self.labeled_indices, self.test_indices = create_dataset_split(
                self.X,
                self.y,
                test_split_seed,
                setting.setting_test_size,
                train_split_seed,
                setting.setting_labeled_train_size,
                setting.setting_train_type,
                setting.factor,
            )

    def get_scenario_id(self):
        """
        Get the scenario id.
        """
        return self.scenario_id

    def get_openml_id(self):
        """
        Get the openml id.
        """
        return self.openml_id

    def get_setting(self):
        """
        Get the setting.
        """
        return self.setting

    def get_seed(self):
        """
        Get the seed.
        """
        return self.seed

    def get_labeled_instances(self):
        """
        Get the labeled instances.
        """
        return self.labeled_indices

    def get_test_indices(self):
        """
        Get the test indices.
        """
        return self.test_indices

    def get_labeled_train_data(self):
        """
        Get the labeled training data.
        """
        return self.X[self.labeled_indices], self.y[self.labeled_indices]

    def get_unlabeled_train_data(self):
        """
        Get the unlabeled training data (X and y).
        """
        combined_train_labeled_test = self.labeled_indices + self.test_indices
        mask = np.array([True] * len(self.X))
        mask[combined_train_labeled_test] = False
        return self.X[mask], self.y[mask]

    def get_test_data(self):
        """
        Get the test data.
        """
        return self.X[self.test_indices], self.y[self.test_indices]

    def get_data_split(self):
        """
        Get labeled, unlabeled and test data.
        """
        X_l, y_l = self.get_labeled_train_data()
        X_u, y_u = self.get_unlabeled_train_data()
        X_test, y_test = self.get_test_data()
        return X_l, y_l, X_u, y_u, X_test, y_test

    def __repr__(self):
        params = dict(self.__dict__)
        params.pop("X")
        params.pop("y")
        return "<ActiveLearningScenario> " + str(params)



class Oracle:
    """Oracle

    This class has access to the true labels of the data and can be used to query the true labels of the data.

    Args:
        X_u (np.ndarray): The unlabeled data.
        y_u (np.ndarray): The true labels of the unlabeled data.

    Attributes:
        X_u (np.ndarray): The unlabeled data.
        y_u (np.ndarray): The true labels of the unlabeled data.
    """

    def __init__(self, X_u=None, y_u=None):
        self.X_u = X_u
        self.y_u = y_u

    def set_data(self, X_u, y_u):
        """Initializes the data."""
        self.X_u = X_u
        self.y_u = y_u

    def query(self, indices_to_label):
        """Queries the true labels of the data.

        Parameters:
            indices_to_label (np.ndarray): The indices of the instances to query.
        Returns:
            np.ndarray: The true labels of the queried instances.
        """
        return self.y_u[indices_to_label]



import numpy as np
import pytest
from pytorch_tabnet.tab_model import TabNetClassifier
from tabpfn.scripts.transformer_prediction_interface import TabPFNClassifier

from alpbench.util.ensemble_constructor import Ensemble


def eval_ensemble_with_learner(learner):
    X, y = np.random.random((10, 10)), np.random.randint(0, 2, 10)
    learner.fit(X, y)
    ens = Ensemble(learner, 10, 10)
    ens.fit(X, y)
    ens.predict(X)


@pytest.mark.usefixtures("benchmark_connector")
def test_init_ensemble(benchmark_connector):
    for bench_learner in benchmark_connector.learners:
        learner = benchmark_connector.load_learner_by_name(bench_learner["learner_name"])
        eval_ensemble_with_learner(learner)


@pytest.mark.usefixtures("benchmark_connector")
def test_init_tabpfn_ensemble(benchmark_connector):
    learner = TabPFNClassifier()
    eval_ensemble_with_learner(learner)


@pytest.mark.usefixtures("benchmark_connector")
def test_init_tabpfn_ensemble_larger_dataset(benchmark_connector):
    learner = TabPFNClassifier()
    X, y = np.random.random((1200, 10)), np.random.randint(0, 2, 1200)
    learner.fit(X[0:999], y[0:999])
    ens = Ensemble(learner, 10, 10)
    ens.fit(X, y)
    ens.predict(X)


@pytest.mark.usefixtures("benchmark_connector")
def test_init_tabnet_ensemble(benchmark_connector):
    learner = TabNetClassifier(verbose=0)
    eval_ensemble_with_learner(learner)



import numpy as np

from alpbench.pipeline.Initializer import RandomInitializer

NUM_SAMPLES = 2


def test_random_initializer():
    rand_init = RandomInitializer(42)
    X_u = np.random.random((10, 10))
    sample = rand_init.sample(X_u=X_u, num_samples=NUM_SAMPLES)
    assert len(sample) == NUM_SAMPLES



import pytest

from alpbench.benchmark.BenchmarkConnector import DataFileBenchmarkConnector
from alpbench.benchmark.Observer import PrintObserver
from alpbench.evaluation.experimenter.DefaultSetup import ensure_default_setup
from alpbench.pipeline.ALPEvaluator import ALPEvaluator

base_folder = "temp/"
learner_file = base_folder + "learner.json"
query_strategy_file = base_folder + "query_strategy.json"
scenario_file = base_folder + "scenario.json"
setting_file = base_folder + "setting.json"


@pytest.fixture
@pytest.mark.usefixtures("scenario")
def benchmark_connector(scenario):
    bc = DataFileBenchmarkConnector(
        learner_file=learner_file,
        query_strategy_file=query_strategy_file,
        scenario_file=scenario_file,
        setting_file=setting_file,
    )
    ensure_default_setup(bc)
    bc.load_or_create_scenario(
        scenario.openml_id,
        test_split_seed=scenario.test_split_seed,
        train_split_seed=scenario.train_split_seed,
        seed=scenario.seed,
        setting_id=scenario.setting.setting_id,
    )
    bc.dump()
    return DataFileBenchmarkConnector(
        learner_file=learner_file,
        query_strategy_file=query_strategy_file,
        scenario_file=scenario_file,
        setting_file=setting_file,
    )


@pytest.fixture
@pytest.mark.usefixtures("benchmark_connector")
def alpevaluator(benchmark_connector):
    return ALPEvaluator(benchmark_connector)


def test_with_learner(alpevaluator):
    alpevaluator.with_learner("rf")
    assert alpevaluator.learner_name == "rf"


def test_with_query_strategy(alpevaluator):
    alpevaluator.with_query_strategy("random")
    assert alpevaluator.query_strategy_name == "random"


def test_with_setting(alpevaluator):
    alpevaluator.with_setting("small")
    assert alpevaluator.setting_name == "small"


def test_with_test_split_seed(alpevaluator):
    alpevaluator.with_test_split_seed(1337)
    assert alpevaluator.test_split_seed == 1337


def test_with_train_split_seed(alpevaluator):
    alpevaluator.with_train_split_seed(42)
    assert alpevaluator.train_split_seed == 42


def test_with_openml_id(alpevaluator):
    alpevaluator.with_openml_id(31)
    assert alpevaluator.openml_id == 31


def test_with_observer(alpevaluator):
    observer = PrintObserver(None, None)
    alpevaluator.with_observer(observer)
    assert observer in alpevaluator.observer_list


def test_get_test_data(alpevaluator):
    alpevaluator.with_setting("small")
    alpevaluator.with_openml_id(31)
    alpevaluator.get_test_data()


def test_fit(alpevaluator):
    alpevaluator.with_setting("small")
    alpevaluator.with_openml_id(31)
    alpevaluator.with_learner("rf_entropy")
    alpevaluator.with_query_strategy("random")
    alpevaluator.fit()



import os
import sqlite3

import numpy as np
import pandas as pd

from alpbench.benchmark.BenchmarkSuite import OpenMLBenchmarkSuite
from alpbench.evaluation.analysis.get_results import StudyDataFromFile
from alpbench.evaluation.analysis.plot_functions import BudgetPerformancePlot
from alpbench.evaluation.experimenter.Experimenter import run

# create config file that specifies the databasename, tables name, and keyfields
content = """
PY_EXPERIMENTER:
  n_jobs: 1

  Database:
    provider: sqlite
    database: ALPBenchmark
    table:
      name: results
      keyfields:
        "setting_name":
          type: VARCHAR(50)
        "openml_id":
          type: INT
        "learner_name":
          type: VARCHAR(50)
        "query_strategy_name":
          type: VARCHAR(50)
        "test_split_seed":
          type: INT
        "train_split_seed":
          type: INT
        "seed":
          type: INT

    logtables:
      accuracy_log:
        model_dict: TEXT

      labeling_log:
        data_dict: TEXT
"""

# Create config directory if it does not exist
if not os.path.exists("config"):
    os.mkdir("config")

# Create config file
experiment_configuration_file_path = os.path.join("config", "exp_config.yml")
with open(experiment_configuration_file_path, "w") as f:
    f.write(content)
db_spec = "config/exp_config.yml"
# load dataset ids from the benchmark suite
cc18 = OpenMLBenchmarkSuite(99)
# we only take the first 3 for demonstration purposes
cc18_ids = cc18.get_openml_dataset_ids()[:3]
setting_combinations = [{"setting_name": "small"}]
parameters = {
    "learner_name": ["rf_entropy", "svm_rbf"],
    "query_strategy_name": ["random", "margin"],
    "test_split_seed": np.arange(1),
    "train_split_seed": np.arange(1),
    "seed": np.arange(1),
    "openml_id": cc18_ids,
}


# check whether the database has been created
created = False
done = False


def initialization():

    global created
    if not created:

        run(
            db_spec,
            run_setup=True,
            reset_experiments=False,
            setting_combinations=setting_combinations,
            parameters=parameters,
        )
        # Specify the path to your .db file
        db_path = "ALPBenchmark.db"

        # Connect to the database
        conn = sqlite3.connect(db_path)

        # Create a cursor object to interact with the database
        cursor = conn.cursor()

        # Get the list of tables in the database
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = cursor.fetchall()

        # Display the contents of each table
        for table_name in tables[:1]:
            table_name = table_name[0]
            print(f"Contents of table {table_name}:")
            query = f"SELECT * FROM {table_name}"
            results_df = pd.read_sql_query(query, conn)

        assert len(results_df) == 12
        created = True

        for index, row in results_df.iterrows():
            if row["status"] != "created":
                created = False

        assert created is True


def experiments():

    global created
    global done

    if not created:
        initialization()

    if not done:
        run(db_spec, run_setup=False, reset_experiments=False)

        # Specify the path to your .db file
        db_name = "ALPBenchmark"
        db_path = db_name + ".db"
        # db_path = "ALPBenchmark.db"
        # Connect to the database
        conn = sqlite3.connect(db_path)

        # Create a cursor object to interact with the database
        cursor = conn.cursor()

        # Get the list of tables in the database
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = cursor.fetchall()

        # Display the contents of each table
        for table_name in tables[:1]:
            table_name = table_name[0]
            print(f"Contents of table {table_name}:")
            query = f"SELECT * FROM {table_name}"
            results_df = pd.read_sql_query(query, conn)

        done = True

        for index, row in results_df.iterrows():
            if row["status"] != "done":
                done = False

        assert done is True


def test_plotting_utilities():

    global done
    if not done:
        experiments()

    # Specify the path to your .db file
    db_name = "ALPBenchmark"
    db_path = db_name + ".db"

    study = StudyDataFromFile("DATAFRAMES", db_path)

    study.get_dataframes()

    dataframe = study.generate_summary_df()
    study.generate_aubc_df(dataframe)

    openmlids = dataframe["openml_id"].unique()
    learners = dataframe["learner_name"].unique()

    for oid in openmlids:
        for learner in learners:
            perfPlot = BudgetPerformancePlot(dataframe, oid, learner, "test_accuracy", "FIGURES/")
            perfPlot.generate_plot_data()
            perfPlot.show(show_fig=False)



from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier

from alpbench.benchmark.Observer import PrintObserver


def test_performance_evaluation():
    X, y = load_iris(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=42)
    model = DecisionTreeClassifier()
    model.fit(X_train, y_train)
    performance_observer = PrintObserver(X_test=X_test, y_test=y_test)
    performance_observer.observe_model(model=model, iteration=0)



import codecs
import os

import setuptools

NAME = "alpbench"
DESCRIPTION = "Active Learning Pipelines Benchmark"
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"
URL = "https://github.com/ValentinMargraf/ActiveLearningPipelines"
EMAIL = "valentin.margraf@ifi.lmu.de"
AUTHOR = "Valentin Margraf et al."
REQUIRES_PYTHON = ">=3.10.0"

work_directory = os.path.abspath(os.path.dirname(__file__))


# https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path):
    with codecs.open(str(os.path.join(work_directory, rel_path)), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delimiter = '"' if '"' in line else "'"
            return line.split(delimiter)[1]


with open(os.path.join(work_directory, "README.md"), encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(work_directory, "CHANGELOG.md"), encoding="utf-8") as f:
    changelog = f.read()

base_packages = [
    "numpy",
    "scipy",
    "matplotlib",
    "pandas",
    "py-experimenter",
    "mysql-connector-python",
    "openml",
    "scikit-learn",
    "scikit-activeml",
    "catboost",
    "xgboost",
]

full_packages = ["pytorch-tabnet", "tabpfn"]


setuptools.setup(
    name=NAME,
    version=get_version("alpbench/__init__.py"),
    description=DESCRIPTION,
    long_description="\n\n".join([readme, changelog]),
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    project_urls={
        "Tracker": "https://github.com/ValentinMargraf/ActiveLearningPipelines/issues",
        "Source": "https://github.com/ValentinMargraf/ActiveLearningPipelines",
        "Documentation": "https://activelearningpipelines.readthedocs.io/",
    },
    packages=setuptools.find_packages(include=("alpbench", "alpbench.*")),
    install_requires=base_packages,
    extras_require={"full": base_packages + full_packages},
    include_package_data=True,
    license="MIT",
    license_files=("LICENSE",),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords=["python", "machine learning", "active learning", "benchmark", "tabular data", "classification"],
    zip_safe=True,
)



import numpy as np

from alpbench.pipeline.Oracle import Oracle


def test_oracle_setter():
    oracle = Oracle()

    X_u = np.random.random((10, 10))
    y_u = np.random.randint(0, 2, 10)
    oracle.set_data(X_u, y_u)

    assert (oracle.X_u == X_u).all()
    assert (oracle.y_u == y_u).all()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.colors import Normalize
import pandas as pd
from py_experimenter.experimenter import PyExperimenter
import json
import os
from scipy.stats import ttest_ind_from_stats
from scipy.stats import ttest_ind as tt


class StudyData():

    def __init__(self):

    def get_aubc_df(self):
        pass

class StudyDataFromFile(StudyData):




class StudyDataFromDatabase(StudyData):
    def __init__(self, source, grid):
        self.db_config, self.scenario_config = source
        self.grid = grid
        self.setting = self.grid["setting_name"]

    def get_aubc_df(self):

        db_name = self.grid["db_name"]
        experimenter_scenarios = PyExperimenter(experiment_configuration_file_path=self.scenario_config,
                                                database_credential_file_path=self.db_config)
        results = experimenter_scenarios.get_table()
        results = results.rename(columns={'ID': 'experiment_id'})
        accuracies = experimenter_scenarios.get_logtable('accuracy_log')
        experiment_ids = results['experiment_id'].values
        print(results['learner_name'].unique())
        print(results['query_strategy_name'].unique())

        labeling = experimenter_scenarios.get_logtable('labeling_log')



        files_exist = False
        if os.path.exists("DATAFRAMES/"+ self.setting + "/" + db_name +"_"+ self.setting +  "_accuracies.csv"):
            files_exist = True
            print("FILE EXISTS")
            loaded_accuracies = pd.read_csv("DATAFRAMES/"+ self.setting + "/" + db_name +"_"+ self.setting + "_accuracies.csv")
            loaded_labeling = pd.read_csv("DATAFRAMES/"+ self.setting + "/" + db_name +"_"+ self.setting + "_labeling.csv")
            loaded_results = pd.read_csv("DATAFRAMES/"+ self.setting + "/"+ db_name +"_"+self.setting + "_results.csv")
            pre_df = pd.read_csv("DATAFRAMES/" + self.setting + "/" + db_name +"_"+ self.setting + "_dynamic.csv")


        if not os.path.exists("DATAFRAMES/" + self.setting + "/"):
            os.makedirs("DATAFRAMES/" + self.setting + "/")

        results.to_csv("DATAFRAMES/"+ self.setting + "/" + db_name +"_"+ self.setting +  "_results.csv")
        accuracies.to_csv("DATAFRAMES/"+ self.setting + "/" + db_name +"_"+self.setting +  "_accuracies.csv")
        labeling.to_csv("DATAFRAMES/"+ self.setting + "/" + db_name +"_"+self.setting +  "_labeling.csv")


        data = accuracies['model_dict'].values
        max_iter = 20

        exp_ids = []
        for i in experiment_ids[:]:
            if i in labeling['experiment_id'].values and i in accuracies['experiment_id'].values:
                if files_exist:
                    if i not in loaded_labeling['experiment_id'].values and i not in loaded_accuracies['experiment_id'].values:
                        exp_ids.append(i)
                else:
                    exp_ids.append(i)

        dict_to_fill = pd.DataFrame(columns=['experiment_id', 'iteration', 'test_accuracy', 'test_f1',
                                                'test_precision', 'test_recall', 'test_auc', 'test_log_loss',
                                                'len_X_l'])

        for enum,i in enumerate(exp_ids[:]):
            if enum%100 == 0:
                print("Experiment", enum, "of", len(exp_ids))
            for j in range(int(max_iter) + 1):
                lab = labeling[labeling['experiment_id'] == i]
                label = lab['data_dict'].values
                actual_max_iter = max(map(int, json.loads(data[0]).keys()))
                if j > actual_max_iter:
                    continue
                label_dict = json.loads(label[0])[str(j)]
                acc = accuracies[accuracies['experiment_id'] == i]
                data = acc['model_dict'].values
                dict = json.loads(data[0])[str(j)]
                dict_to_fill.loc[len(dict_to_fill)] = [int(i), int(dict['iteration']), dict['test_accuracy'],
                                                        dict['test_f1'],
                                                        dict['test_precision'], dict['test_recall'],
                                                        dict['test_auc'], dict['test_log_loss'],
                                                        int(label_dict['len_X_l'])]

        merged_df = pd.merge(results, dict_to_fill, on='experiment_id', how='inner')
        merged_df = merged_df[['setting_name', 'openml_id', 'learner_name', 'query_strategy_name',
                                'test_split_seed', 'train_split_seed', 'seed', 'iteration', 'test_accuracy',
                                'test_f1',
                                'test_precision', 'test_recall', 'test_auc', 'test_log_loss', 'len_X_l']]
        end_df = merged_df

        if files_exist:
            end_df = pd.concat([pre_df, end_df], axis=0)
        return end_df
(


import numpy as np
from sklearn.ensemble import RandomForestClassifier

from alpbench.benchmark.Observer import PrintObserver
from alpbench.evaluation.experimenter.LogTableObserver import LogTableObserver, SparseLogTableObserver


class MockupResultProcessor:
    def __init__(self):
        pass

    def process_logs(self, data):
        pass


def test_labeling_statistics():
    X, y = np.random.random((10, 10)), np.random.randint(0, 2, (10, 1))
    po = PrintObserver(X, y)
    po.observe_data(1, X, y, X, y, X)

    po.observe_data(1, np.empty((10, 0)), np.empty((1, 0)), X, y, X)


def test_model_stats():
    X, y = np.random.random((10, 10)), np.random.randint(0, 2, (10, 1))
    po = PrintObserver(X, y)
    rf = RandomForestClassifier()
    rf.fit(X, y)
    po.observe_model(1, rf)


def test_logtableobserver():
    X, y = np.random.random((10, 10)), np.random.randint(0, 2, (10, 1))
    lto = LogTableObserver(MockupResultProcessor(), X, y)
    rf = RandomForestClassifier()
    rf.fit(X, y)
    lto.observe_data(1, X, y, X, y, X)
    lto.observe_model(1, rf)


def test_sparselogtableobserver():
    X, y = np.random.random((10, 10)), np.random.randint(0, 2, (10, 1))
    lto = SparseLogTableObserver(MockupResultProcessor(), X, y)
    rf = RandomForestClassifier()
    rf.fit(X, y)
    lto.observe_data(1, X, y, X, y, X)
    lto.observe_model(1, rf)
    lto.log_data({})
    lto.log_model({})



import pytest

from alpbench.benchmark.ActiveLearningScenario import create_dataset_split

from .fixtures.scenario import LABELED_INDICES, LEN_DATASET, OPENML_ID, SCENARIO_ID, SEED, SETTING_ID, TEST_INDICES


@pytest.mark.usefixtures("scenario")
def test_get_setting_id(scenario):
    assert scenario.get_scenario_id() == SCENARIO_ID


@pytest.mark.usefixtures("scenario")
def test_get_openml_id(scenario):
    assert scenario.get_openml_id() == OPENML_ID


@pytest.mark.usefixtures("scenario")
def test_get_test_data(scenario):
    X, y = scenario.get_test_data()
    assert len(X) == len(TEST_INDICES)
    assert len(X) == len(TEST_INDICES)


@pytest.mark.usefixtures("scenario")
def test_get_labeled_train_data(scenario):
    X, y = scenario.get_labeled_train_data()
    assert len(X) == len(LABELED_INDICES)
    assert len(X) == len(LABELED_INDICES)


@pytest.mark.usefixtures("scenario")
def test_get_unlabeled_train_data(scenario):
    X, y = scenario.get_unlabeled_train_data()
    exp_len = LEN_DATASET - len(LABELED_INDICES) - len(TEST_INDICES)
    assert len(X) == exp_len
    assert len(y) == exp_len


@pytest.mark.usefixtures("scenario")
def test_get_seed(scenario):
    assert scenario.get_seed() == SEED


@pytest.mark.usefixtures("scenario")
def test_get_setting(scenario):
    setting = scenario.get_setting()
    assert str(type(setting)) == "<class 'alpbench.benchmark.ActiveLearningSetting.ActiveLearningSetting'>"
    assert setting.get_setting_id() == SETTING_ID


@pytest.mark.usefixtures("scenario")
def test_create_dataset_split(scenario):
    labeled_indices, test_indices = create_dataset_split(
        scenario.X,
        scenario.y,
        scenario.test_split_seed,
        scenario.setting.setting_test_size,
        scenario.train_split_seed,
        scenario.setting.setting_labeled_train_size,
        scenario.setting.setting_train_type,
        scenario.setting.factor,
    )
    assert len(labeled_indices) > 0
    assert len(test_indices) > 0
    labeled_indices, test_indices = create_dataset_split(
        scenario.X,
        scenario.y,
        scenario.test_split_seed,
        scenario.setting.setting_test_size,
        scenario.train_split_seed,
        scenario.setting.setting_labeled_train_size,
        scenario.setting.setting_train_type,
        3,
    )
    assert len(labeled_indices) > 0
    assert len(test_indices) > 0

    labeled_indices, test_indices = create_dataset_split(
        scenario.X,
        scenario.y,
        scenario.test_split_seed,
        scenario.setting.setting_test_size,
        scenario.train_split_seed,
        2,
        "absolute",
        scenario.setting.factor,
    )
    assert len(labeled_indices) > 0
    assert len(test_indices) > 0

    labeled_indices, test_indices = create_dataset_split(
        scenario.X,
        scenario.y,
        scenario.test_split_seed,
        scenario.setting.setting_test_size,
        scenario.train_split_seed,
        2,
        "absolute",
        2,
    )



from abc import ABC, abstractmethod

import openml


class BenchmarkSuite(ABC):
    """Benchmark Suite

    Abstract class for benchmark suites. A benchmark suite is a collection of datasets from OpenML.

    Args:
        name (str): name of the benchmark suite

    Attributes:
        name (str): name of the benchmark suite
    """

    def __init__(self, name="TabZilla"):
        self.name = name

    @abstractmethod
    def get_openml_dataset_ids(self):
        """
        Abstract method for getting the OpenML dataset ids of the benchmark suite.
        """
        pass


class TabZillaBenchmarkSuite(BenchmarkSuite):
    """TabZilla Benchmark Suite

    This benchmark suite contains the dataset ids of the TabZilla benchmark suite
    (https://github.com/naszilla/tabzilla).

    Args:
        None

    Attributes:
        name (str): name of the benchmark suite
    """

    def __init__(self):
        super().__init__(name="TabZilla")

    def get_openml_dataset_ids(self):
        """
        Get the OpenML dataset ids of the TabZilla benchmark suite.

        Returns:
            list: list of OpenML dataset ids
        """
        return [
            11,
            14,
            22,
            25,
            29,
            31,
            46,
            51,
            54,
            151,
            334,
            470,
            846,
            934,
            1043,
            1067,
            1169,
            1459,
            1468,
            1486,
            1489,
            1493,
            1494,
            1567,
            4134,
            4538,
            23512,
            40536,
            40981,
            41027,
            41143,
            41147,
            41150,
            41159,
        ]


class OpenMLBenchmarkSuite(BenchmarkSuite):
    """OpenML Benchmark Suite

    This benchmark suite contains the dataset ids of the OpenML-CC18 benchmark suite
    (https://openml.github.io/openml-python/main/examples/20_basic/simple_suites_tutorial.html).

    Args:
        openml_benchmark_id (int): id of the OpenML benchmark suite

    Attributes:
        openml_benchmark_id (int): id of the OpenML benchmark suite
        name (str): name of the benchmark suite
        openml_dataset_ids (list): list of OpenML dataset ids
    """

    def __init__(self, openml_benchmark_id, name="OpenML-BenchmarkSuite"):
        super(OpenMLBenchmarkSuite, self).__init__(name=name + "-" + str(openml_benchmark_id))
        self.openml_benchmark_id = openml_benchmark_id

        benchmark = openml.study.get_suite(openml_benchmark_id)
        tasks = openml.tasks.list_tasks(task_id=benchmark.tasks)

        self.openml_dataset_ids = list()
        for k, t in tasks.items():
            self.openml_dataset_ids.append(t["did"])

    def get_openml_dataset_ids(self):
        """
        Get the OpenML dataset ids of the OpenML-CC18 benchmark suite.

        Returns:
            list: list of OpenML dataset ids
        """
        return self.openml_dataset_ids
